# Configuration file for the Sphinx documentation builder.
# Templated from both RTD tutorial and doped by SMTG-Bham.

# -- Project information

project = 'findTDE'
copyright = '2023, Alexander S. Hauck'
author = 'Alexander S. Hauck'

release = '1.0'
version = '1.1.0'

# -- General configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosectionlabel',
    'sphinx_book_theme',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme'

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = "findtde-logo.png"
html_title = "findTDE"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

html_theme_options = {
    "repository_url": "https://github.com/hauck-as/findTDE",
    "github_repo": "https://github.com/hauck-as/findTDE",
    "github_button": True,
    "github_user": "hauck-as", # Username
    "description": "Set of scripts to facilitate easy calculations of threshold displacement energies for materials in VASP/LAMMPS using ab initio/classical molecular dynamics.",
    "repository_branch": "main",
    "path_to_docs": "docs",
    "use_repository_button": True,
    "home_page_in_toc": True,
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org",
        "colab_url": "https://colab.research.google.com",
    },
}

# Adding “Edit Source” links on your Sphinx theme
html_context = {
    "display_github": True, # Integrate GitHub
    "github_user": "hauck-as", # Username
    "github_repo": "findTDE", # Repo name
    "github_version": "main", # Version
    "conf_py_path": "/docs/", # Path in the checkout to the docs root
}

# -- Options for EPUB output
epub_show_urls = 'footnote'



# Copyright (c) 2023 Alexander Hauck. Distributed under the terms of the MIT License.

"""
findTDE comprises a set of scripts to facilitate easy, high-throughput calculations of threshold displacement energies for materials in VASP/LAMMPS using ab initio/classical molecular dynamics.
"""

__author__ = "Alexander Hauck"
__copyright__ = "Copyright (C) 2023 Alexander Hauck"
__license__ = "MIT License"
__version__ = "1.1.0"



#!/usr/bin/env python3
"""Python module used to submit several find_tde jobs at once."""
import os
import re
import subprocess
import math as m
import numpy as np

base_path = os.getcwd()
bin_path, inp_path, perfect_path = os.path.relpath('bin', base_path), os.path.relpath('inp', base_path), os.path.relpath('perfect', base_path)


def write_find_tde_calcs(direction, atom_type, atom_number, ke_i=25, ke_cut=40, mode='L', directions_filepath=os.path.relpath('latt_dirs_to_calc.csv', os.getcwd())):
    """
    Writes the text file used by find_tde to perform calculations. Requires a direction list (L or S)
    as well as the atom type (string) and number (int), uses 25 eV initial KE and lattice direction
    mode by default (can use any integer KE or spherical direction mode (S)).
    """
    lat_dir_list_header_text = '########################################\n' + \
    '# format of text file\n' + \
    '# nL    atom_type    atom_number    ke_i    ke_cut    u    v    w\n' + \
    '# n+1S    atom_type    atom_number    ke_i    ke_cut    r    p    t\n' + \
    '########################################'
    
    # if file does exist, append directions to it starting from previous number
    if os.path.isfile(directions_filepath):
        ld_f = open(directions_filepath, 'r+')
        print('Appending to file')
        n, n_all = 0, []
        ld_f_lines = ld_f.readlines()
        for line in ld_f_lines:
            line = line.strip('\n')
            if line[0] != '#':
                n_all.append(re.split(r'[;,\s]\s*', line)[0][:-1])
                dir_n = re.split(r'[;,\s]\s*', line)[5:]
                if np.all(np.array(dir_n, dtype=float) == direction.astype(float)):
                    print('Direction matches previous pseudo')
                    n = re.split(r'[;,\s]\s*', line)[0][:-1]
        n_prev_max = int(np.amax(np.array((n_all), dtype=int)))
        if n != 0:
            ld_f.write(', '.join(['\n'+str(n)+str(mode), str(atom_type), str(atom_number), str(ke_i), str(ke_cut), str(direction[0]), str(direction[1]), str(direction[2])]))
        else:
            ld_f.write(', '.join(['\n'+str(1+n_prev_max)+str(mode), str(atom_type), str(atom_number), str(ke_i), str(ke_cut), str(direction[0]), str(direction[1]), str(direction[2])]))
        ld_f.close()
    
    # if file does not exist, create file with header and add direction to it
    else:
        ld_f = open(directions_filepath, 'w+')
        print('Creating new file')
        ld_f.write(lat_dir_list_header_text)
        ld_f.write(', '.join(['\n'+str(1)+str(mode), str(atom_type), str(atom_number), str(ke_i), str(ke_cut), str(direction[0]), str(direction[1]), str(direction[2])]))
        ld_f.close()
    
    return


def find_multiple_tde(directions, atom_type, atom_number, ke_i=25, ke_cut=40, mode='L', conv='standard', prog='vasp', lmp_ff='AlGaN.sw', screen_num=0):
    """
    Function used to write and submit multiple instances for find_tde.
    """
    for i in range(directions.shape[0]):
        write_find_tde_calcs(directions[i, :], atom_type, atom_number, ke_i=ke_i, ke_cut=ke_cut, mode=mode)
        # os.system('screen -S TDE'+str(i)+' find_tde -c '+str(conv))
        if prog == 'vasp':
            subprocess.run(['screen', '-S', 'TDE'+str(screen_num+i)+atom_type.lower()+str(atom_number), 'find_tde', '-c', str(conv), '-p', str(prog)], capture_output = True, text = True)
        elif prog == 'lammps':
            subprocess.run(['screen', '-S', 'TDE'+str(screen_num+i)+atom_type.lower()+str(atom_number), 'find_tde', '-c', str(conv), '-p', str(prog), '-f', str(lmp_ff)], capture_output = True, text = True)
    return


def lat2sph(uvw, ai):
    """
    Function converting from lattice directions to spherical coordinates. Given two 2D
    arrays, one for lattice directions and one for lattice vectors, returns a 2D array
    of the spherical coordinates corresponding to the lattice directions.
    """
    xyz, rpt = np.zeros(uvw.shape), np.zeros(uvw.shape)
    
    for i in range(uvw.shape[0]):
        xyz[i, :] = uvw[i, :]@ai
        rpt[i, :] = np.around(cart2sph(xyz[i, 0], xyz[i, 1], xyz[i, 2]), decimals=2)
        rpt[i, 0] = 1.
    
    return rpt


def scale_C3z_symmetry(rpt):
    """
    Function to scale all given spherical coordinates to a single zone based on threefold
    rotation symmetry about the c-axis. Given a 2D array of spherical coordinates, returns
    another 2D array of spherical coordinates with polar angle scaled by 120 deg increments
    to the desired range.
    """
    for i in range(rpt.shape[0]):
        if rpt[i, 1] >= 30. and rpt[i, 1] <= 150.:
            pass
        elif rpt[i, 1] >= 0. and rpt[i, 1] < 30.:
            rpt[i, 1] += 120.
        elif rpt[i, 1] > 150. and rpt[i, 1] <= 270.:
            rpt[i, 1] -= 120.
        elif rpt[i, 1] > 270. and rpt[i, 1] <= 360.:
            rpt[i, 1] -= 240.
        else:
            raise Exception('Angle outside of 0-360 deg')
    return rpt


def write_sph_dirs(rpt, directions_filepath=os.path.relpath('sph_directions.csv', os.getcwd())):
    """
    Function to write spherical directions to a text/csv file for multi_tde.py use.
    Given a 2D array of spherical coordinates, writes array values to the file
    represented by the filepath string.
    """
    sph_f_header = '########################################\n' + \
    'Spherical directions for multi_tde.py calculations\n' + \
    'r, p, t\n' + \
    '########################################\n'
    
    sph_f = open(directions_filepath, 'a+')
    if os.stat(directions_filepath).st_size == 0:
        sph_f.write(sph_f_header)
    sph_f.write('########################################\n')
    for i in range(rpt.shape[0]):
        sph_line = ', '.join([str(round(j, 2)) for j in rpt[i, :]])
        sph_f.write(sph_line+'\n')
    sph_f.close()
    return


def rewrite_find_tde_calcs_keys(base_path=os.getcwd(), directions_filename='latt_dirs_to_calc.csv', ke_cut=45):
    """
    Writes the text file used by find_tde to perform calculations. Requires a direction list (L or S)
    as well as the atom type (string) and number (int), uses 25 eV initial KE and lattice direction
    mode by default (can use any integer KE or spherical direction mode (S)).
    """
    directions_filepath=os.path.join(base_path, directions_filename)
    data_files = glob.glob(base_path+'*/*_out.txt', recursive=True)
    
    lat_dir_list_header_text = '########################################\n' + \
    '# format of text file\n' + \
    '# nL    atom_type    atom_number    ke_i    ke_cut    u    v    w\n' + \
    '# n+1S    atom_type    atom_number    ke_i    ke_cut    r    p    t\n' + \
    '########################################'
    
    if os.path.isfile(directions_filepath):
        lat_dir_list_lines = []
    else:
        lat_dir_list_lines = [lat_dir_list_header_text]
    
    for i in range(len(data_files)):
        pseudo = re.search(r'\d+\w', subprocess.run(['grep', 'Lattice Direction Pseudo:', data_files[i]], capture_output = True, text = True).stdout.split('\n')[0]).group()
        atom_type = subprocess.run(['grep', 'Atom....................:', data_files[i]], capture_output = True, text = True).stdout.split('\n')[0].split(': ')[-1]
        atom_number = re.search(r'\d+', subprocess.run(['grep', 'Atom Number.............:', data_files[i]], capture_output = True, text = True).stdout).group()
        ke_i = re.search(r'\d+', subprocess.run(['grep', 'KE:', data_files[i]], capture_output = True, text = True).stdout.split('\n')[0]).group()
        v_direction = re.split(r'[\s]\s*', subprocess.run(['grep', 'direction:', data_files[i]], capture_output = True, text = True).stdout.split('\n')[0].split(': ')[-1].replace('[', '').replace(']', '').strip())
        pseudo_line = ', '.join(['\n'+pseudo, atom_type, atom_number, ke_i, str(ke_cut), str(v_direction[0]), str(v_direction[1]), str(v_direction[2])])
        lat_dir_list_lines.append(pseudo_line)
    
    ld_f = open(directions_filepath, 'a+')
    for line in lat_dir_list_lines:
        ld_f.write(line)
    ld_f.close()
    
    return


def split_pseudo_atom_info(pseudo_atoms, pseudo_keyfile, savefile=False):
    """
    Function to split pseudo_atom info and determine direction associated with pseudo.
    Returns array with information.
    """
    # empty array for findTDE run info
    converted_pa_info = np.zeros(pseudo_atoms.shape[0], dtype=[('direction', list), ('atom_type', '<U4'), ('atom_num', '<i4'), ('dir_mode', 'U1')])
    
    # read in pseudo keys from file into dict
    pseudo_keys = pseudo_keys_from_file(pseudo_keyfile=pseudo_keyfile)
    
    for i in range(pseudo_atoms.shape[0]):
        # split pseudo_atoms
        pseudo, atom_info = pseudo_atoms[i].rsplit('_', 1) # string for pseudo and atom type/number
        atom_type, atom_num = re.search(r'\D+', atom_info).group(), int(re.search(r'\d+', atom_info).group())    # specified atom type and number for findTDE run
        dir_mode = pseudo[-1]
        
        pseudo_direction = ((1.,) + tuple(pseudo_keys[pseudo]))
        converted_pa_info[i] = (pseudo_direction, atom_type, atom_num, dir_mode)
    
    if savefile == False:
        pass
    elif type(savefile) == str:
        np.savetxt(savefile, converted_pa_info, delimiter=';', fmt=['%s', '%s', '%d', '%s'])
    else:
        raise TypeError('Savefile name must be a valid string')
    
    return converted_pa_info


def read_pseudo_atom_info(pseudo_atom_filename):
    """
    Read pseudo_atom info into structured numpy array to be used for running findTDE.
    """
    # read in info into numpy array from txt file
    pa_info_arr = np.genfromtxt(pseudo_atom_filename, delimiter=';', dtype=[('direction', tuple), ('atom_type', '<U4'), ('atom_num', '<i4'), ('dir_mode', 'U1')])
    
    # decode direction tuple byte string into tuple
    for i in range(pa_info_arr.shape[0]):
        pa_info_arr[i]['direction'] = eval(pa_info_arr[i]['direction'].decode('utf-8'))
    
    return pa_info_arr


# generate directions for calculations
"""l_directions = np.array([
    [-1, 4, 1],
    [-1, 7, 3],
    [-1, 8, 4],
    [-2, -1, 0],
    [-4, 11, 0],
    [-4, 4, 3],
    [0, 0, -1],
    [0, 0, 1],
    [0, 1, -5],
    [0, 1, 0],
    [0, 2, 1],
    [0, 4, 1],
    [0, 5, 2],
    [1, -1, 1],
    [1, 1, 1],
    [1, 2, -5],
    [1, 2, 0],
    [1, 7, 0],
    [2, -2, 1],
    [2, 0, -5],
    [2, 1, 0],
    [2, 4, -5],
    [4, -3, 0],
    [4, -5, 4],
    [4, 1, 3],
    [4, 3, 1],
    [5, 2, 1],
    [5, 3, 0],
    [7, 0, 3],
    [7, 4, 5],
    [8, 4, 3],
    [1, 1, 0],
    [-1, -1, 0],
    [-1, 1, 0]
])

s_directions = np.array([
    # [1., 30., 105.],
    # [1., 90., 60.]
    [1., 90., 97.5]
])"""

"""
ga_s_gan_dirs = np.array([
    [1., 45., 112.5],
    [1., 30., 105.],
    [1., 90., 60.]
])

ga_l_gan_dirs = np.array([
    [5, 2, 1],
    [-4, 4, 3],
    [-2, -1, 0]
])

n_s_gan_dirs = np.array([
    [1., 90., 112.5],
    [1., 90., 97.5]
])

n_l_gan_dirs = np.array([
    [-1, 1, 0]
])

al_s_aln_dirs = np.array([
    [1., 82.5, 52.5],
    [1., 82.5, 60.],
    [1., 82.5, 67.5],
    [1., 82.5, 45.]
])

n_s_aln_dirs = np.array([
    [1., 90., 45.],
    # [1., 90., 112.5]
])
"""

# write_sph_dirs(scale_C3z_symmetry(lat2sph(all_tde_data[1], gan_lattice_vecs)), filepath=os.path.relpath('sph_directions.csv', base_path))
# sph_directions = np.genfromtxt(os.path.relpath('sph_directions.csv', base_path), delimiter=',', comments='#', skip_header=4)
sph_directions_ext_s1 = np.genfromtxt(os.path.relpath('sph_directions_ext_set1.csv', base_path), delimiter=',', comments='#', skip_header=4)
# sph_directions_ext_s2 = np.genfromtxt(os.path.relpath('sph_directions_ext_set2.csv', base_path), delimiter=',', comments='#', skip_header=4)
# sph_directions_ext_s3 = np.genfromtxt(os.path.relpath('sph_directions_ext_set3.csv', base_path), delimiter=',', comments='#', skip_header=4)
## aln_redo_info = read_pseudo_atom_info(os.path.join(base_path, 'aln_lmp_errs_run_info.txt'))
## for i in range(aln_redo_info.shape[0]):
##     find_multiple_tde(np.array([aln_redo_info[i]['direction']]), aln_redo_info[i]['atom_type'], aln_redo_info[i]['atom_num'], ke_i=10, ke_cut=100, mode=aln_redo_info[i]['dir_mode'], conv='midpoint', prog='lammps', lmp_ff='AlGaN.sw', screen_num=30)

# write directions and run find_tde
# ga 34, n 35
# find_multiple_tde(s_directions, 'ga', 34, ke_i=25, ke_cut=45, mode='S', conv='midpoint', prog='vasp')

## find_multiple_tde(sph_directions, 'zn', 34, ke_i=25, ke_cut=45, mode='S', conv='midpoint', prog='vasp', screen_num=100)
## find_multiple_tde(sph_directions, 'o', 35, ke_i=25, ke_cut=45, mode='S', conv='midpoint', prog='vasp', screen_num=500)
find_multiple_tde(sph_directions_ext_s1, 'zn', 34, ke_i=10, ke_cut=200, mode='S', conv='midpoint', prog='lammps', lmp_ff='ZnO.tersoff', screen_num=1010)

# calculate lowest energy directions from GaN/AlN calcs, both Ga/Al & N knockouts, in VASP
# find_multiple_tde(ga_s_gan_dirs, 'al', 34, ke_i=25, ke_cut=45, mode='S', conv='midpoint', prog='vasp', screen_num=4410)
# find_multiple_tde(ga_l_gan_dirs, 'al', 34, ke_i=25, ke_cut=45, mode='L', conv='midpoint', prog='vasp', screen_num=4420)
# find_multiple_tde(n_s_gan_dirs, 'n', 35, ke_i=25, ke_cut=45, mode='S', conv='midpoint', prog='vasp', screen_num=4430)
# find_multiple_tde(n_l_gan_dirs, 'n', 35, ke_i=25, ke_cut=45, mode='L', conv='midpoint', prog='vasp', screen_num=4440)
# find_multiple_tde(al_s_aln_dirs, 'al', 34, ke_i=25, ke_cut=45, mode='S', conv='midpoint', prog='vasp', screen_num=4450)
# find_multiple_tde(n_s_aln_dirs, 'n', 35, ke_i=25, ke_cut=45, mode='S', conv='midpoint', prog='vasp', screen_num=4460)

# calculate extended spherical direction mesh in LAMMPS
# find_multiple_tde(sph_directions_ext_s3, 'ga', 34, ke_i=10, ke_cut=100, mode='S', conv='midpoint', prog='lammps', lmp_ff='AlGaN.sw', screen_num=1100)
# find_multiple_tde(sph_directions_ext_s3, 'n', 35, ke_i=10, ke_cut=100, mode='S', conv='midpoint', prog='lammps', lmp_ff='AlGaN.sw', screen_num=1400)

# calculate lowest energy directions from GaN/AlN calcs, both Ga/Al & N knockouts, in LAMMPS
# find_multiple_tde(ga_s_gan_dirs, 'al', 34, ke_i=10, ke_cut=100, mode='S', conv='midpoint', prog='lammps', lmp_ff='AlGaN.sw', screen_num=7010)
# find_multiple_tde(ga_l_gan_dirs, 'al', 34, ke_i=10, ke_cut=100, mode='L', conv='midpoint', prog='lammps', lmp_ff='AlGaN.sw', screen_num=7020)
# find_multiple_tde(n_s_gan_dirs, 'n', 35, ke_i=10, ke_cut=100, mode='S', conv='midpoint', prog='lammps', lmp_ff='AlGaN.sw', screen_num=7030)
# find_multiple_tde(n_l_gan_dirs, 'n', 35, ke_i=10, ke_cut=100, mode='L', conv='midpoint', prog='lammps', lmp_ff='AlGaN.sw', screen_num=7040)
# find_multiple_tde(al_s_aln_dirs, 'al', 34, ke_i=10, ke_cut=100, mode='S', conv='midpoint', prog='lammps', lmp_ff='AlGaN.sw', screen_num=7050)
# find_multiple_tde(n_s_aln_dirs, 'n', 35, ke_i=10, ke_cut=100, mode='S', conv='midpoint', prog='lammps', lmp_ff='AlGaN.sw', screen_num=7060)


#!/usr/bin/env python3
"""Python module used to analyze find_tde calculations."""
import os
import glob
from pathlib import Path

import subprocess
import re
import fortranformat as ff
import pprint

import numpy as np
import pandas as pd

import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots
import plotly.io as pio
# import seaborn as sns

from findtde.utils import *
from findtde.plotting import *
from findtde.io import *

base_path = Path.cwd()
bin_path, inp_path, perfect_path = base_path / 'bin', base_path / 'inp', base_path / 'perfect'


def vasp_inp_match_check(poscar_filepath, potcar_filepath):
    """
    Function to compare the order of elements in the POSCAR file to the order of
    elements in the POTCAR file and ensure the order matches.
    """
    # open POSCAR file and read lines into a list
    pos_f = open(poscar_filepath, 'r')
    pos_lines = pos_f.readlines()
    pos_f.close()
    
    # convert species names line from POSCAR file from Fortran to a list 
    pos_spec_line = ff.FortranRecordReader('(2A5)')
    pos_species = pos_lines[5]
    pos_spec_list = pos_spec_line.read(pos_species)
    for i in range(len(pos_spec_list)):
        pos_spec_list[i] = pos_spec_list[i].strip()
    
    # open POTCAR file and read lines into a list
    pot_f = open(potcar_filepath, 'r')
    pot_lines = pot_f.readlines()
    pot_f.close()
    
    # find POTCAR lines with species names
    pot_spec_lines = subprocess.run(['grep', 'TITEL', potcar_filepath], capture_output = True, text = True)
    pot_spec_lines_list = pot_spec_lines.stdout.split('\n ')
    pot_spec_list = []
    for i in range(len(pot_spec_lines_list)):
        pot_spec_list.append(re.split(r'[;,\s]\s*', pot_spec_lines_list[i])[4])
    
    if pos_spec_list == pot_spec_list:
        return True
    else:
        return False


# function to analyze find_tde runs and determine if there are any issues
def check_find_tde_runs(tde_calc_dir=base_path, program='vasp', ke_tol=1, temp_tol=0.6):
    """
    Function used to check if TDE calcs have proper velocity given to the correct atom, correct KE, if the final temperature is proper relative to initial temperature.
    Input is a directory path string containing find_tde calculations, and function returns a dictionary of runs with errors.
    """
    # empty dicts for final checked simulations and found errors
    all_find_tde_checks_dict, all_find_tde_errs_dict = {}, {}
    
    # find each overall findTDE run directory
    data_filepaths = np.array(glob.glob(os.path.join(tde_calc_dir, '*', '*_data.csv'), recursive=True))
    
    # determine values for a set of findTDE runs corresponding to a given data file
    for i in range(data_filepaths.shape[0]):
        pseudo_path = data_filepaths[i].rsplit('/', 1)[0]    # relative path for findTDE pseudo
        # vasp runs use pseudo_atominfo, lammps runs have an additional _lmp after
        if program == 'vasp':
            pseudo_atom = pseudo_path.rsplit('/', 1)[-1]    # string for pseudo and atom type/number
        elif program == 'lammps':
            pseudo_atom = pseudo_path.rsplit('/', 1)[-1].rsplit('_', 1)[0]    # string for pseudo and atom type/number
        pseudo, atom_info = pseudo_atom.rsplit('_', 1)    # separate pseudo/atom info
        atom_type_ideal, atom_num_ideal = re.search(r'\D+', atom_info).group(), int(re.search(r'\d+', atom_info).group())    # specified atom type and number for findTDE run
        find_tde_checks_dict = {
            'tde': [],
            'knockout': [],
            'ke': [],
            'temp': []
                             }
        
        # check if TDE was found from last line of outfile
        out_filepath = glob.glob(os.path.join(pseudo_path, '*_out.txt'))[0]
        with open(out_filepath) as of:
            for line in of:
                pass
            last_line = line
        
        if last_line.startswith('Loop complete, TDE found:'):
            find_tde_checks_dict['tde'].append(True)
        elif last_line.startswith('Loop not complete, TDE above cutoff:'):
            find_tde_checks_dict['tde'].append('Cutoff')
        else:
            find_tde_checks_dict['tde'].append(False)
            
        # checking values for a VASP run
        if program == 'vasp':
            for subdir, dirs, files in os.walk(pseudo_path):
                for k in range(len(dirs)):
                    if dirs[k] == 'cgm':    # ignore post-AIMD CGM runs
                        continue
                    else:
                        ke_calc_path = os.path.join(subdir, dirs[k])

                        # compare specified KE from directory to the actual initial value from the OUTCAR
                        # ke_tol gives tolerance in eV, necessary due to momentum correction in VASP
                        ke_i_ideal = float(re.search(r'\d+', dirs[k]).group())
                        ke_i_actual = float(re.search(r'-?\d*\.{0,1}\d+', subprocess.run(['grep', 'EKIN', os.path.join(ke_calc_path, 'OUTCAR')], capture_output = True, text = True).stdout.strip('\n').split('\n')[0]).group())
                        if (ke_i_actual >= ke_i_ideal - ke_tol) and (ke_i_actual <= ke_i_ideal + ke_tol):
                            find_tde_checks_dict['ke'].append(True)
                        else:
                            find_tde_checks_dict['ke'].append(False)

                        # determine final to initial lattice temperature ratio, should be roughly half or less due to equipartition theorem of thermodynamics
                        # temp_tol gives tolerance value for ratio, halving of temperature corresponds to 0.5
                        temps = subprocess.run(['grep', 'EKIN_LAT', os.path.join(ke_calc_path, 'OUTCAR')], capture_output = True, text = True).stdout.strip('\n').split('\n')
                        temp_i = float(re.search(r'\d*\.{0,1}\d+', temps[0].split('(')[-1]).group())
                        temp_f = float(re.search(r'\d*\.{0,1}\d+', temps[-1].split('(')[-1]).group())
                        tfti = temp_f/temp_i
                        if tfti < temp_tol:
                            find_tde_checks_dict['temp'].append(True)
                        else:
                            find_tde_checks_dict['temp'].append(False)

                        # determine from initial POSCAR if velocity vector is given to correct atom
                        pos_f = open(os.path.join(ke_calc_path, 'POSCAR'), 'r')
                        pos_lines = pos_f.readlines()
                        pos_f.close()

                        # convert species names line from POSCAR file from Fortran to a list 
                        spec_line = ff.FortranRecordReader('(2A5)')
                        species = pos_lines[5]
                        spec_list = spec_line.read(species)
                        for i in range(len(spec_list)):
                            spec_list[i] = spec_list[i].strip()

                        # convert ions per species line from POSCAR file from Fortran to a list
                        no_line = ff.FortranRecordReader('(2I6)')
                        ion_nos = pos_lines[6]
                        no_list = no_line.read(ion_nos)
                        total_ions = sum(no_list)

                        # convert velocity vector line from POSCAR file from Fortran format to a list
                        vel_line = ff.FortranRecordWriter('(3E16.8)')
                        for i in range(len(spec_list)):           
                            if atom_type_ideal.lower() in spec_list[i].lower():
                                vel_vector = pos_lines[8+total_ions+atom_num_ideal+sum(no_list[:i])].strip('\n')

                        # check if velocity vector for the desired knockout atom is nonzero
                        if vel_vector == vel_line.write(np.array([0, 0, 0])):
                            find_tde_checks_dict['knockout'].append(False)
                        elif vel_vector != vel_line.write(np.array([0, 0, 0])):
                            find_tde_checks_dict['knockout'].append(True)

        elif program == 'lammps':
            pass
        
        # create summary of errors found and detailed checks done
        all_find_tde_errs_dict[pseudo_atom] = []
        if False in find_tde_checks_dict['tde']:
            all_find_tde_errs_dict[pseudo_atom].append('tde')
        if False in find_tde_checks_dict['knockout']:
            all_find_tde_errs_dict[pseudo_atom].append('knockout')
        if False in find_tde_checks_dict['ke']:
            all_find_tde_errs_dict[pseudo_atom].append('ke')
        if False in find_tde_checks_dict['temp']:
            all_find_tde_errs_dict[pseudo_atom].append('temp')
        if not all_find_tde_errs_dict[pseudo_atom]:
            all_find_tde_errs_dict.pop(pseudo_atom)
        
        all_find_tde_checks_dict[pseudo_atom] = find_tde_checks_dict
    
    return (all_find_tde_errs_dict, all_find_tde_checks_dict)


# create dataframes for final energies and energy differences
def find_tde_analysis(atom_types, atom_nums, datafile='all_tde_data.csv', keyfile=os.path.join(base_path, 'latt_dirs_to_calc.csv')):
    """
    Function to analyze gathered find_tde data and format appropriately. Inputs are
    two lists for each atom type (strings) and atom numbers (integers) for which data
    is desired. Returns a tuple with two dictionaries. The first dictionary contains
    tuples with each dictionary key being the atom type and number concatenated and the
    tuples containing two pandas DataFrames. The first DataFrame is the final energy for
    each calculation direction at each calculated kinetic energy, and the second DataFrame
    is the same for final energy difference from the perfect crystal energy. The second
    dictionary contains keys for the lattice direction pseudos.
    """
    find_tde_data = np.genfromtxt(datafile, delimiter=', ', skip_header=1, dtype=str)
    find_tde_key = np.genfromtxt(keyfile, delimiter=', ', skip_header=4, dtype=str)
    
    ##### need to return the cutoff energy as well as the lattice directions #####
    
    # create overall dictionary for each instance of atom types and numbers
    all_df_dict, pseudo_keys = {}, {}
    
    # find minimum cutoff energy to use for analysis
    try:
        ke_cutoff = np.amin(find_tde_key[:, 4].astype('int'))
    except IndexError:
        ke_cutoff = np.amin(find_tde_key[4].astype('int'))
    
    # sort data by atom types, atom numbers, and calculation directions
    for i in range(len(atom_types)):
        for j in range(len(atom_nums)):
            atom_indices = np.where((find_tde_data[:, 1] == atom_types[i].lower()) & (find_tde_data[:, 2] == str(atom_nums[j])))[0]
            toten_dict, toten_dif_dict = {}, {}
            for c, k in enumerate(atom_indices):
                # determine directions from pseudos and group data by pseudo/direction
                if find_tde_data[k, 0] == find_tde_data[atom_indices[c-1], 0]:
                    try:
                        toten_dict[find_tde_data[k, 0]].update({int(find_tde_data[k, 3]):float(find_tde_data[k, 4])})
                        toten_dif_dict[find_tde_data[k, 0]].update({int(find_tde_data[k, 3]):float(find_tde_data[k, 5])})
                    except KeyError:
                        if len(np.where(find_tde_data[k, 0] == find_tde_key[0])) == 1:
                            if find_tde_data[k, 0][-1] == 'L':
                                latdir = find_tde_key[-3:]
                            elif find_tde_data[k, 0][-1] == 'S':
                                latdir = find_tde_key[-2:]
                        pseudo_keys[find_tde_data[k, 0]] = latdir
                        toten_dict[find_tde_data[k, 0]], toten_dif_dict[find_tde_data[k, 0]] = {int(find_tde_data[k, 3]):float(find_tde_data[k, 4])}, {int(find_tde_data[k, 3]):float(find_tde_data[k, 5])}
                elif find_tde_data[k, 0] != find_tde_data[atom_indices[c-1], 0]:
                    if find_tde_data[k, 0][-1] == 'L':
                        latdir = find_tde_key[np.where(find_tde_data[k, 0] == find_tde_key[:, 0])[0][0], :][-3:]
                    elif find_tde_data[k, 0][-1] == 'S':
                        latdir = find_tde_key[np.where(find_tde_data[k, 0] == find_tde_key[:, 0])[0][0], :][-2:]
                    pseudo_keys[find_tde_data[k, 0]] = latdir
                    toten_dict[find_tde_data[k, 0]], toten_dif_dict[find_tde_data[k, 0]] = {int(find_tde_data[k, 3]):float(find_tde_data[k, 4])}, {int(find_tde_data[k, 3]):float(find_tde_data[k, 5])}

            # add to overall dictionary if data has been found
            if toten_dict and toten_dif_dict:
                # dataframe from dictionary with final TOTEN values
                toten_df = pd.DataFrame(data=toten_dict)
                toten_df.sort_index(inplace=True)

                # dataframe from dictionary with calculated differences from perfect crystal TOTEN
                toten_dif_df = pd.DataFrame(data=toten_dif_dict)
                toten_dif_df.sort_index(inplace=True)
            
                # add dataframes 
                all_df_dict[atom_types[i].lower()+str(atom_nums[j])] = (toten_df, toten_dif_df)
    
    return (all_df_dict, pseudo_keys)


def find_tde_sph_analysis(find_tde_dE, pseudo_keys, lattice_vecs, e_tol=1.0, ke_cut=45):
    """
    Given a dataframe with lattice direction pseudo columns, initial KE indices, and delta E
    values, returns a dictionary listing this info and the knockout's spherical coordinates.
    """
    find_tde_sph_dict = {}
    
    for i in range(find_tde_dE.columns.shape[0]):
        tde_info_dict = {}
        for j in range(find_tde_dE.index.shape[0]):
            if np.isnan(find_tde_dE[find_tde_dE.columns[i]].loc[find_tde_dE.index[j]]) == True:
                continue
            elif find_tde_dE[find_tde_dE.columns[i]].loc[find_tde_dE.index[j]] > e_tol:
                if find_tde_dE.index[j] <= ke_cut:
                    tde_info_dict['TDE'] = find_tde_dE.index[j]
                    tde_info_dict['dE'] = find_tde_dE[find_tde_dE.columns[i]].loc[find_tde_dE.index[j]]
                elif find_tde_dE.index[j] > ke_cut:
                    tde_info_dict['TDE'] = ke_cut
                break
            else:
                tde_info_dict['TDE'] = ke_cut
            
            if find_tde_dE.columns[i][-1] == 'S':
                tde_info_dict['phi'], tde_info_dict['theta'] = pseudo_keys[find_tde_dE.columns[i]].astype(float)
            elif find_tde_dE.columns[i][-1] == 'L':
                tde_info_dict['latt_dir'] = pseudo_keys[find_tde_dE.columns[i]]
                rho, tde_info_dict['phi'], tde_info_dict['theta'] = lat2sph(np.reshape(pseudo_keys[find_tde_dE.columns[i]], (1, 3)).astype(int), lattice_vecs)[0]
        
        find_tde_sph_dict[find_tde_dE.columns[i]] = tde_info_dict
    
    return find_tde_sph_dict


def generate_tde_sph_arr(tde_data_df, pseudo_keys, lattice_vecs, e_tol=1.0, ke_cut=45, polar_offset=60):
    """
    Function to reorganize dataframe into a Nx3 array with spherical coordinates.
    """
    find_tde_sph_dict = find_tde_sph_analysis(tde_data_df, pseudo_keys=pseudo_keys, lattice_vecs=lattice_vecs, e_tol=e_tol, ke_cut=ke_cut)
    find_tde_pseudos = find_tde_sph_dict.keys()
    find_tde_sph_arr = np.zeros((len(find_tde_pseudos), 3))
    for k in range(len(find_tde_pseudos)):
        key = list(find_tde_pseudos)[k]
        try:
            rpt = np.array([[1.0, find_tde_sph_dict[key]['phi'], find_tde_sph_dict[key]['theta']]])
            rpt_scaled = scale_C6z_symmetry(scale_C3z_symmetry(rpt))    # scale_C3z_symmetry(rpt)
            find_tde_sph_arr[k, :] = np.array([
                # 60 DEG ADDED TO POLAR ANGLE SO THAT 0 deg ALIGNS TO THE RIGHT IN VISUALIZATION
                (round(rpt_scaled[0, 1], 2)+polar_offset),
                round(rpt_scaled[0, 2], 2),
                round(find_tde_sph_dict[key]['TDE'], 2)
            ])
        except:
            pass
    find_tde_sph_arr = find_tde_sph_arr[~np.all(find_tde_sph_arr == 0, axis=1)]
    
    return (find_tde_sph_arr, find_tde_pseudos)


# function to gather data from find_tde lammps runs, recalculated after thermalizing
def tde_thrm_data_gather(ofile='all_tde_data_lmp_thrm.csv', tde_calc_dir=base_path, e_tol=1.0):
    """
    Gathers all TDE data formatted as find_tde data.
    """
    tde_data_header = 'Lattice Direction Pseudo, Atom Type, Atom Number, KE, Final Energy, Delta Energy\n'
    thrm_data_header = 'Lattice Direction Pseudo_Atom TypeAtom Number, KE, Calcs with Defects, Total Calcs\n'
    out_filepaths = np.array(glob.glob(os.path.join(tde_calc_dir, '*/*_out.txt'), recursive=True))
    
    for i in range(out_filepaths.shape[0]):
        pseudo_path = out_filepaths[i].rsplit('/', 1)[0]
        pseudo_atom = pseudo_path.rsplit('/', 1)[-1].rsplit('_', 1)[0]
        pseudo_datafile = os.path.join(pseudo_path, pseudo_atom + '_data.csv')
        pf = open(pseudo_datafile, 'w+')
        pf.write(tde_data_header)
        
        for subdir, dirs, files in os.walk(pseudo_path):
            for k in range(len(dirs)):
                ke_calc_path = os.path.join(subdir, dirs[k])
                ke = re.search(r'\d+', dirs[k]).group()
                # Ei from log_thermal.lammps may be different depending on size
                # Ei = re.search(r'-?\d*\.{0,1}\d+', subprocess.run(['grep', 'Initial energy of atoms', os.path.join(ke_calc_path, 'log_thermal.lammps')], capture_output = True, text = True).stdout.strip('\n').split('\n')[-1]).group()
                log_filepaths = np.array(glob.glob(os.path.join(ke_calc_path, 'log_*.lammps'), recursive=True))
                
                for j in range(log_filepaths.shape[0]):
                    if log_filepaths[j].rsplit('/', 1)[-1] == 'log_thermal.lammps':
                        pass
                    else:
                        pseudo, atom_info = pseudo_atom.rsplit('_', 1)
                        atom_type, atom_num = re.search(r'\D+', atom_info).group(), re.search(r'\d+', atom_info).group()
                        Ei = re.search(r'-?\d*\.{0,1}\d+', subprocess.run(['grep', 'Initial energy of atoms', log_filepaths[j]], capture_output = True, text = True).stdout.strip('\n').split('\n')[-1]).group()
                        Ef = re.search(r'-?\d*\.{0,1}\d+', subprocess.run(['grep', 'Final energy of atoms', log_filepaths[j]], capture_output = True, text = True).stdout.strip('\n').split('\n')[-1]).group()
                        dE = float(Ef) - float(Ei)
                        data_line = pseudo + ', ' + atom_type + ', ' + atom_num + ', ' + ke + ', ' + Ef + ', ' + str(dE) + '\n'
                        pf.write(data_line)
        pf.close()
    
    
    data_filepaths = np.array(glob.glob(os.path.join(tde_calc_dir, '*/*_data.csv'), recursive=True))
    tde_thrm_dict = {}
    
    for file in data_filepaths:
        pseudo_data = np.genfromtxt(file, delimiter=', ', skip_header=1,
                                    dtype=[('f0', '<U4'), ('f1', '<U4'), ('f2', '<i4'),
                                           ('f3', '<i4'), ('f4', '<f8'), ('f5', '<f8')]
                                   )
        pseudo_atom = pseudo_data[0][0] + '_' + pseudo_data[0][1] + str(pseudo_data[0][2])
        tde_thrm_dict[pseudo_atom] = np.array([[pseudo_data[0][3], 0, 0]], dtype=int)
        
        for i in range(pseudo_data.shape[0]):
            if pseudo_data[i][3] not in tde_thrm_dict[pseudo_atom][:, 0]:
                tde_thrm_dict[pseudo_atom] = np.append(tde_thrm_dict[pseudo_atom], np.array([[pseudo_data[i][3], 0, 0]]), axis=0)
            for j in range(tde_thrm_dict[pseudo_atom].shape[0]):
                if pseudo_data[i][3] == tde_thrm_dict[pseudo_atom][j, 0]:
                    tde_thrm_dict[pseudo_atom][j, 2] += 1
                    if pseudo_data[i][-1] >= e_tol:
                        tde_thrm_dict[pseudo_atom][j, 1] += 1
                    elif pseudo_data[i][-1] < e_tol:
                        pass
                elif pseudo_data[i][3] != tde_thrm_dict[pseudo_atom][j, 0]:
                    pass
    
    of = open(tde_calc_dir+ofile, 'w+')
    of.write(thrm_data_header)
    of_lines = []
    
    for psatan in tde_thrm_dict.keys():
        for n in range(tde_thrm_dict[psatan].shape[0]):
            of_lines.append(psatan + ', ' + str(tde_thrm_dict[psatan][n, 0]) + ', ' + str(tde_thrm_dict[psatan][n, 1]) + ', ' + str(tde_thrm_dict[psatan][n, 2]) + '\n')
    for line in of_lines:
        of.write(line)
    
    of.close()
    
    return tde_thrm_dict


def evaluate_tdes(knockout_atom_type, knockout_atom_num, run_type='vasp', base_path=base_path, annotate_pseudos=False, interpolate_heatmap=False):
    """
    Function to execute all analysis and plotting functions for find_tde calculations.
    """
    knockout_atom = knockout_atom_type.lower() + str(knockout_atom_num)

    # read in VASP and LAMMPS perfect structure information
    if run_type == 'vasp':
        vasp_ref_file = base_path / 'inp' / 'POSCAR'
    elif run_type == 'lammps':
        vasp_ref_file, lmp_ref_file = base_path / 'inp' / 'POSCAR', base_path / 'inp' / 'read_data_perfect.lmp'
    pos_f = open(vasp_ref_file, 'r')
    pos_lines = pos_f.readlines()
    pos_f.close()

    # open POSCAR file and read lattice vector lines into a list
    ai = [pos_lines[2][1:-1], pos_lines[3][1:-1], pos_lines[4][1:-1]]

    # convert lattice vectors line from POSCAR file from Fortran to a list
    lat_vec_line = ff.FortranRecordReader('(3F22.16)')
    vasp_lattice_vecs = np.array([lat_vec_line.read(ai[j]) for j in range(len(ai))])
    vasp_params = np.array([np.linalg.norm(vasp_lattice_vecs[0, :]), np.linalg.norm(vasp_lattice_vecs[1, :]), np.linalg.norm(vasp_lattice_vecs[2, :])])

    # check if runs have issues
    # should change from print, maybe try except?
    print(check_find_tde_runs(tde_calc_dir=base_path, program=run_type)[0])

    ######## TDE data ########
    if run_type == 'vasp':
        # gather TDE data from calculations 
        tde_data_gather(ofile=(base_path / 'all_tde_data.csv'), tde_calc_dir=base_path)
        
        # organize data from csv file into dataframes and plot lineplot
        all_find_tde_dfs, pseudo_keys = find_tde_analysis([knockout_atom_type.capitalize()], [knockout_atom_num], datafile=(base_path / 'all_tde_data.csv'), keyfile=(base_path / 'latt_dirs_to_calc.csv'))
        find_tde_df = all_find_tde_dfs[knockout_atom][1]

        generate_tde_line_plot(find_tde_df, im_write=True, im_name=(base_path / 'tde_lineplot.png'))
        
        # reorganize data into array and plot scatter plot
        tde_sph_arr, tde_pseudos = generate_tde_sph_arr(find_tde_df, pseudo_keys, lattice_vecs=vasp_lattice_vecs, e_tol=1.0, ke_cut=45, polar_offset=angle_between([1., 0., 0.], vasp_lattice_vecs[0]))
        
        generate_tde_scatter_plot(tde_sph_arr, tde_pseudos, txt_show=annotate_pseudos, im_write=True, im_name=(base_path / 'tde_scatter.png'))

        # from Victor, read data into a (nsamples x 3) array (x, y, f(x, y)), interpolate and plot heatmap data
        if interpolate_heatmap:
            ps_tde, ts_tde, es_tde = idw_heatmap(tde_sph_arr, RES=tde_sph_arr.shape[0], P=5)

            generate_tde_heatmap_plot(ps_tde, ts_tde, es_tde, im_write=True, im_name=(base_path / 'tde_heatmap.png'))
        
    elif run_type == 'lammps':
        # gather TDE data from calculations 
        tde_data_gather(ofile=(base_path / 'all_tde_data_lmp.csv'), tde_calc_dir=base_path)
        
        # organize data from csv file into dataframes and plot lineplot
        all_find_tde_lmp_dfs, lmp_pseudo_keys = find_tde_analysis([knockout_atom_type.capitalize()], [knockout_atom_num], datafile=(base_path / 'all_tde_data_lmp.csv'), keyfile=(base_path / 'latt_dirs_to_calc.csv'))
        find_tde_lmp_df = all_find_tde_lmp_dfs[knockout_atom][1]

        generate_tde_line_plot(find_tde_lmp_df, im_write=True, im_name=(base_path / 'tde_lineplot.png'))
        
        # reorganize data into array and plot scatter plot
        tde_lmp_sph_arr, tde_lmp_pseudos = generate_tde_sph_arr(find_tde_lmp_df, lmp_pseudo_keys, lattice_vecs=vasp_lattice_vecs, e_tol=1.0, ke_cut=100, polar_offset=angle_between([1., 0., 0.], vasp_lattice_vecs[0]))
        
        generate_tde_scatter_plot(tde_lmp_sph_arr, tde_lmp_pseudos, txt_show=annotate_pseudos, im_write=True, im_name=(base_path / 'tde_scatter.png'))

        # from Victor, read data into a (nsamples x 3) array (x, y, f(x, y)), interpolate and plot heatmap data
        if interpolate_heatmap:
            ps_lmp_tde, ts_lmp_tde, es_lmp_tde = idw_heatmap(tde_lmp_sph_arr, RES=tde_lmp_sph_arr.shape[0], P=5)

            generate_tde_heatmap_plot(ps_lmp_tde, ts_lmp_tde, es_lmp_tde, im_write=True, im_name=(base_path / 'tde_heatmap.png'))
    
    return



# -*- coding: utf-8 -*-
# Copyright (c) 2024 Alexander Hauck. Distributed under the terms of the MIT License.
import sys
import argparse
import warnings
from pathlib import Path

from findtde import __version__
from findtde.cli.main_functions import *
import pydefect.cli.main as pyd_parse


description = """findTDE comprises a set of scripts to facilitate easy,
high-throughput calculations of threshold displacement energies for materials
in VASP/LAMMPS using ab initio/classical molecular dynamics."""


epilog = f'Author: Alexander Hauck Version: {__version__}'


def parse_args_main(args):
    parser = argparse.ArgumentParser(description=description, epilog=epilog)
    subparsers = parser.add_subparsers()

    return parser.parse_args(args)


def main():
    args = parse_args_main(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()


#!/usr/bin/env python3
"""Python module of utility functions for findTDE calculations."""
from math import gcd
from fractions import Fraction
import numpy as np


# math functions
def unit_vector(vector):
    # https://stackoverflow.com/a/13849249
    """ Returns the unit vector of the vector.  """
    return vector / np.linalg.norm(vector)


def angle_between(v1, v2):
    # https://stackoverflow.com/a/13849249
    """ Returns the angle in radians between vectors 'v1' and 'v2'::

            >>> angle_between((1, 0, 0), (0, 1, 0))
            1.5707963267948966
            >>> angle_between((1, 0, 0), (1, 0, 0))
            0.0
            >>> angle_between((1, 0, 0), (-1, 0, 0))
            3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))*(180/np.pi)


def cart2sph(x, y, z):
    dxy = np.sqrt(x**2 + y**2)
    r = np.sqrt(dxy**2 + z**2)
    theta = np.arctan2(y, x)
    phi = np.arctan2(dxy, z)
    theta, phi = np.rad2deg([theta, phi])
    return r, theta % 360, phi


def sph2cart(theta, phi, r=0.5):
    theta, phi = np.deg2rad([theta, phi])
    z = r * np.cos(phi)
    rsinphi = r * np.sin(phi)
    x = rsinphi * np.cos(theta)
    y = rsinphi * np.sin(theta)
    return x, y, z


# crystallography functions
# https://ssd.phys.strath.ac.uk/resources/crystallography/crystallographic-direction-calculator/
def lattice_iconv(indices, ind_type='UVTW'):
    """
    Function converting between rectangular and hexagonal lattice indices. Given a
    tuple of indices and the initial index type as a keyword argument. Index type
    is either [uvw] for rectangular or [UVTW] for hexagonal. Returns a tuple of
    the opposite type.
    """
    print(indices)
    
    if ind_type == 'UVTW':
        U, V, T, W = indices
        u = 2*U + V
        v = 2*V + U
        w = W
        n = gcd(u, gcd(v, w))
        
        u /= n
        v /= n
        w /= n
        
        new_indices = (int(u), int(v), int(w))
    
    elif ind_type == 'uvw':
        u, v, w = indices
        U = Fraction(((2*u) - v), 3)
        V = Fraction(((2*v) - u), 3)
        T = -1*(U + V)
        W = w
        
        denom_gcd = gcd(U.denominator, gcd(V.denominator, T.denominator))
        U *= denom_gcd
        V *= denom_gcd
        T *= denom_gcd
        W *= denom_gcd
        
        new_indices = (int(U), int(V), int(T), int(W))
    
    return new_indices


def lat2sph(uvw, ai):
    """
    Function converting from lattice directions to spherical coordinates. Given two 2D
    arrays, one for lattice directions and one for lattice vectors, returns a 2D array
    of the spherical coordinates corresponding to the lattice directions.
    """
    xyz, rpt = np.zeros(uvw.shape), np.zeros(uvw.shape)
    
    for i in range(uvw.shape[0]):
        xyz[i, :] = uvw[i, :]@ai
        rpt[i, :] = np.around(cart2sph(xyz[i, 0], xyz[i, 1], xyz[i, 2]), decimals=2)
        rpt[i, 0] = 1.00
    
    return rpt


def sph2lat(rpt, ai):
    """
    Function converting from lattice directions to spherical coordinates. Given two 2D
    arrays, one for lattice directions and one for lattice vectors, returns a 2D array
    of the spherical coordinates corresponding to the lattice directions.
    """
    xyz, uvw = np.zeros(rpt.shape), np.zeros(rpt.shape)
    
    for i in range(rpt.shape[0]):
        xyz[i, :] = sph2cart(rpt[i, 1], rpt[i, 2], r=rpt[i, 0])
        uvw[i, :] = xyz[i, :]@np.linalg.inv(ai)
        n = gcd(round(uvw[i, 0]), gcd(round(uvw[i, 1]), round(uvw[i, 2])))
        uvw[i, :] /= n
    
    return uvw


def scale_C3z_symmetry(rpt):
    """
    Function to scale all given spherical coordinates to a single zone based on threefold
    rotation symmetry about the c-axis. Given a 2D array of spherical coordinates, returns
    another 2D array of spherical coordinates with polar angle scaled by 120 deg increments
    to the desired range.
    """
    for i in range(rpt.shape[0]):
        if rpt[i, 1] >= 30. and rpt[i, 1] <= 150.:
            pass
        elif rpt[i, 1] >= 0. and rpt[i, 1] < 30.:
            rpt[i, 1] += 120.
        elif rpt[i, 1] > 150. and rpt[i, 1] <= 270.:
            rpt[i, 1] -= 120.
        elif rpt[i, 1] > 270. and rpt[i, 1] <= 360.:
            rpt[i, 1] -= 240.
        else:
            raise Exception('Angle outside of 0-360 deg')
    return rpt


def scale_C6z_symmetry(rpt):
    """
    Function to scale all given spherical coordinates to a single zone based on sixfold
    rotation symmetry about the c-axis. Given a 2D array of spherical coordinates, returns
    another 2D array of spherical coordinates with polar angle scaled by 60 deg increments
    to the desired range.
    """
    rpt_C3z = scale_C3z_symmetry(rpt)
    
    for i in range(rpt_C3z.shape[0]):
        if rpt_C3z[i, 1] >= 30. and rpt_C3z[i, 1] <= 90.:
            pass
        elif rpt_C3z[i, 1] > 90. and rpt_C3z[i, 1] <= 150.:
            rpt[i, 1] = 180. - rpt[i, 1]
        else:
            raise Exception('Angle outside of 30-150 deg')
    return rpt


# interpolation functions
def idw(samples, tx, ty, P=5):
    """
    Function to compute a single IDW interpolation of sample data.
    Change P value for different interpolation (P>2 is recommended).
    """
    def dist(a, b):
        return ((a[0]-b[0])**2 + (a[1]-b[1])**2)**(1./2.)

    num = 0.
    den = 0.
    for i in range(0, len(samples)):
        d = (dist(samples[i], [tx, ty]))**P
        if(d < 1.e-5):
            return samples[i][2]

        w = 1/d
        num += w*samples[i][2]
        den += w

    return num/den


def idw_heatmap(inputdata, RES=360, P=5):
    """
    Perform full IDW interpolation on a (nsamples x 3) array (x, y, f(x, y))
    and produce data able to be used in a heatmap. From Victor.
    Change P value for different interpolation (P>2 is recommended).
    RES is the resolution of the final image.
    """
    # from Victor, plot heatmap
    # Setup z as a function of interpolated x, y
    minx, maxx = min([d[0] for d in inputdata]), max([d[0] for d in inputdata])
    miny, maxy = min([d[1] for d in inputdata]), max([d[1] for d in inputdata])
    minz, maxz = min([d[2] for d in inputdata]), max([d[2] for d in inputdata])    # useful to scale 0 < z < 1
    dx, dy = (maxx - minx)/(RES-1), (maxy - miny)/(RES-1)
    xs = [minx + i*dx for i in range(0, RES)]
    ys = [miny + i*dy for i in range(0, RES)]
    zs = [[None for i in range(0, RES)] for j in range(0, RES)]
    for i in range(0, RES):
        for j in range(0, RES):
            zs[i][j] = idw(samples=inputdata, tx=xs[i], ty=ys[j], P=P)
    # print(xs, ys, zs)
    return (xs, ys, np.transpose(zs))



#!/usr/bin/env python3
"""Python module of plotting functions for findTDE calculations."""
import os
from pathlib import Path
import random as rand
import numpy as np
import pandas as pd

import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots
import plotly.io as pio
# import seaborn as sns

from findtde.utils import idw, idw_heatmap

base_path = Path.cwd()
bin_path, inp_path, perfect_path = base_path / 'bin', base_path / 'inp', base_path / 'perfect'

# interpolation and plotting functions
txt_positions_reg = ['top center', 'bottom center']
txt_positions_extra = ['top center', 'bottom center', 'top right', 'top left', 'bottom right', 'bottom left']


def improve_text_position(x, txt_positions=txt_positions_reg):
    """
    Function to improve the text position for Plotly scatter plots.
    More efficient if the x values are sorted.
    """
    return [txt_positions[(i % len(txt_positions)-(rand.randint(1,2)))] for i in range(len(x))]


# Plotly hover text formatting for heatmaps
hover_se = '<br><b>Polar</b>: %{x:.2f}'+\
            '<br><b>Azimuthal</b>: %{y:.2f}'+\
            '<br><b>Ef</b>: %{z:.2f} eV'

hover_tde = '<br><b>Polar</b>: %{x:.2f}'+\
            '<br><b>Azimuthal</b>: %{y:.2f}'+\
            '<br><b>TDE</b>: %{color:.2f} eV'


# Change Plotly default template to simple white and modify for 
pl_paper_theme = pio.templates['simple_white']
pl_paper_theme.layout.xaxis.ticks = 'inside'
pl_paper_theme.layout.yaxis.ticks = 'inside'
pl_paper_theme.layout.xaxis.mirror = 'ticks'  # True | "ticks" | False | "all" | "allticks"
pl_paper_theme.layout.yaxis.mirror = 'ticks'  # True | "ticks" | False | "all" | "allticks"
pl_paper_theme.layout.font.size = 32
# pl_paper_theme.layout.xaxis.title.standoff = 20
pl_paper_theme.layout.xaxis.title.font.size = 40
# pl_paper_theme.layout.yaxis.title.standoff = 20
pl_paper_theme.layout.yaxis.title.font.size = 40
#pl_paper_theme.layout.coloraxis.colorbar.title.standoff = 20
pio.templates.default = pl_paper_theme


def generate_tde_line_plot(tde_data_df, im_write=False, im_name='tde_lineplot.png'):
    """
    Given a DataFrame of TDE data, produces line plot for each calculated direction with KE on x-axis
    and difference in final energy on y-axis.
    """
    fig = px.line(tde_data_df, #log_y=True,
                  markers=True, color_discrete_sequence=px.colors.qualitative.Dark24) # Alphabet, Light24, & Dark24 have most

    fig.update_traces(connectgaps=True, marker_size=9, line = dict(width=4, dash='dash'))
    fig.update_layout(autosize=False, width=1300, height=600,
                      xaxis_title=r'$KE_{i} \text{ [eV]}$',
                      yaxis_title=r'$\Delta E \text{ [eV]}$',
                      legend_title='Pseudo',
                      font_size=20
                     )
    fig.show()
    
    if im_write == True:
        fig.write_image(os.path.join(base_path, im_name))
    elif im_write == False:
        pass
    
    return


def generate_tde_scatter_plot(tde_sph_arr, tde_directions, txt_show=False, im_write=False, im_name='tde_scatter_plot.png'):
    """
    Given an Nx3 array of TDE data, produces line plot for each calculated direction with polar angle on x-axis,
    azimuthal angle on y-axis, TDE value as color, and direction pseudos as text.
    """
    fig = go.Figure(data=go.Scatter(x=tde_sph_arr[:, 0],
                                    y=tde_sph_arr[:, 1],
                                    text=list(tde_directions) if txt_show == True else None,
                                    mode='markers+text',
                                    marker=dict(color=tde_sph_arr[:, 2],
                                                colorscale='thermal_r',
                                                size=40, colorbar=dict(thickness=20) #, title=r'$E_{d} \; [eV]$', title_side='right')
                                               )
                                   ))
    
    if txt_show == True:
        fig.update_traces(textposition=improve_text_position(list(tde_directions), txt_positions=txt_positions_reg))
    elif txt_show == False:
        pass

    fig.add_vline(x=90, line=dict(color='black', width=3, dash='dash'))
    fig.add_vline(x=210, line=dict(color='black', width=3, dash='dash'))

    fig.update_layout(# title=('Interpolated SEs from Sph. Pert. of Ga #34 with Latt. Dir. TDEs'),
                      # xaxis_title=r'$\phi \; [^{\circ}]$',
                      # yaxis_title=r'$\theta \; [^{\circ}]$',
                      xaxis=dict(tickmode='linear', dtick=30),
                      yaxis=dict(tickmode='linear', dtick=30),
                      xaxis_range=[80, 160],
                      # xaxis_range=[80, 220],
                      yaxis_range=[0, 180],
                      yaxis_autorange='reversed',
                      # font_size=32,  # 16
                      autosize=False, width = 1200, height = 900  # width = 1600, height = 900
                     )

    fig.show()
    
    if im_write == True:
        fig.write_image(os.path.join(base_path, im_name), scale=2)
    elif im_write == False:
        pass
    
    return


def generate_tde_heatmap_plot(polars, azimuthals, energies, im_write=False, im_name='tde_heatmap.png'):
    """
    Given an Nx3 array of TDE data, produces line plot for each calculated direction with polar angle on x-axis,
    azimuthal angle on y-axis, TDE value as color, and direction pseudos as text.
    """
    fig = go.Figure(data=go.Heatmap(x=polars,
                                    y=azimuthals,
                                    z=energies,
                                    hovertemplate=hover_tde,
                                    colorbar_title=r'$E_{d} \; [eV]$',
                                    colorbar_title_side='right',
                                    colorscale='electric_r'
                                   ))

    fig.add_vline(x=90, line=dict(color='black', width=3, dash='dash'))
    fig.add_vline(x=210, line=dict(color='black', width=3, dash='dash'))

    fig.update_layout(# title=('Interpolated Ga #34 with Latt. Dir. TDEs'),
                      # xaxis_title=r'$\phi \; [^{\circ}]$',
                      # yaxis_title=r'$\theta \; [^{\circ}]$',
                      xaxis=dict(tickmode='linear', dtick=30),
                      yaxis=dict(tickmode='linear', dtick=30),
                      xaxis_range=[90, 210],
                      # xaxis_range=[90, 150],
                      yaxis_range=[0, 180],
                      yaxis_autorange='reversed',
                      font_size=16,  # 24
                      autosize=False, width = 1400, height = 900  # width = 1600, height = 900
                     )
    
    fig.show()
    
    if im_write == True:
        fig.write_image(os.path.join(base_path, im_name))
    elif im_write == False:
        pass
    
    return



#!/usr/bin/env python3
"""Python module of file management functions for findTDE calculations."""
import os
import glob
from pathlib import Path

import numpy as np
import pandas as pd

base_path = Path.cwd()
bin_path, inp_path, perfect_path = base_path / 'bin', base_path / 'inp', base_path / 'perfect'


def find_contcars(tde_calc_dir=base_path, pseudo='*'):
    """
    Function to create an array of post-CGM CONTCAR files for defect analysis.
    """
    contcar_files = np.array(glob.glob(os.path.join(tde_calc_dir, pseudo, '*', 'cgm', 'CONTCAR'), recursive=True))
    return contcar_files


def find_dumpfiles(tde_calc_dir=base_path, pseudo='*'):
    """
    Function to create an array of post-MD dump files for defect analysis.
    """
    dump_files = np.array(glob.glob(os.path.join(tde_calc_dir, pseudo, '*', 'dump.final'), recursive=True))
    return dump_files

def pseudo_keys_to_file(pseudo_keys, pseudo_keyfile=os.path.join(base_path, 'pseudo_keys.csv')):
    """
    Function to convert a pseudo keys dictionary to a CSV.
    """
    pk_df = pd.DataFrame.from_dict(pseudo_keys, orient='index')
    pk_df.to_csv(pseudo_keyfile)
    return


def pseudo_keys_from_file(pseudo_keyfile=(base_path / 'pseudo_keys.csv')):
    """
    Function to convert a CSV file of pseudo keys to a dictionary.
    """
    pk_df = pd.read_csv(pseudo_keyfile, index_col=0)
    pks_dict = pk_df.to_dict('split')
    pks_dict = dict(zip(pks_dict['index'], pks_dict['data']))
    return pks_dict


def write_sph_dirs(rpt, filepath=os.path.join(base_path, 'sph_directions.csv')):
    """
    Function to write spherical directions to a text/csv file for multi_tde.py use.
    Given a 2D array of spherical coordinates, writes array values to the file
    represented by the filepath string.
    """
    sph_f_header = '########################################\n' + \
    'Spherical directions for multi_tde.py calculations\n' + \
    'r, p, t\n' + \
    '########################################\n'
    
    sph_f = open(filepath, 'a+')
    if os.stat(filepath).st_size == 0:
        sph_f.write(sph_f_header)
    sph_f.write('########################################\n')
    for i in range(rpt.shape[0]):
        sph_line = ', '.join([str(round(j, 2)) for j in rpt[i, :]])
        sph_f.write(sph_line+'\n')
    sph_f.close()
    return


# function to gather data from find_tde runs
def tde_data_gather(ofile='all_tde_data.csv', tde_calc_dir=base_path):
    """
    Gathers all TDE data formatted as find_tde data.
    """
    tde_data_header = 'Lattice Direction Pseudo, Atom Type, Atom Number, KE, Final Energy, Delta Energy\n'
    
    data_filepaths = glob.glob(os.path.join(tde_calc_dir, '*', '*_data.csv'), recursive=True)
    of = open(ofile, 'a+')
    of.write(tde_data_header)
    
    for file in data_filepaths:
        data = np.genfromtxt(file, delimiter=', ', skip_header=1,
                             dtype=[('f0', '<U4'), ('f1', '<U4'), ('f2', '<i4'),
                                    ('f3', '<i4'), ('f4', '<f8'), ('f5', '<f8')]
                            )
        np.savetxt(of, data, delimiter=', ', newline='\n',
                   fmt=['%s', '%s', '%d', '%d', '%f', '%f']
                  )
    
    of.close()


# function to determine TDE from a single datafile
def get_tde_from_datafile(data_filepath='*_data.csv', e_tol=1.0, ke_cut=45):
    """
    Scans a findTDE *_data.csv file and returns the TDE value for the specified knockout type.
    """
    # read in tde data into structured array
    tde_data = np.genfromtxt(data_filepath, delimiter=', ', skip_header=1,
                             dtype=[('pseudo', '<U4'), ('atom_type', '<U4'), ('atom_num', '<i4'),
                                ('ke', '<i4'), ('e_fin', '<f8'), ('dE', '<f8')]
                            )
    
    # sort array from lowest KE to highest KE
    tde_data = np.sort(tde_data, order='ke')
    
    # find TDE value from energy difference
    # first KE corresponding to a dE > e_tol is the TDE in the sorted array
    tde_value = 0
    for i in range(tde_data.shape[0]):
        if np.isnan(tde_data[i]['dE']) == True:
            continue
        elif tde_data[i]['dE'] > e_tol:
            if tde_data[i]['ke'] <= ke_cut:
                tde_value = tde_data[i]['ke']
            elif tde_data[i]['ke'] > ke_cut:
                tde_value = ke_cut
            break
        else:
            tde_value = ke_cut
    
    return tde_value



#!/usr/bin/env python3
"""Python module used to read lattice direction and KE text files and generate velocity vectors for VASP."""
import os
import subprocess
import sys
import re
import math as m
import fortranformat as ff
import numpy as np

atom_type, atom_num, lat_dir_pseudo, sim_prog = str(sys.argv[1]), int(sys.argv[2]), str(sys.argv[3]), str(sys.argv[4])

# set directory variables, using PWD as the base directory
base_path = os.getcwd()
bin_path, inp_path, perfect_path = os.path.relpath('bin', base_path), os.path.relpath('inp', base_path), os.path.relpath('perfect', base_path)
if sim_prog == 'vasp':
    tde_run_path = os.path.relpath(lat_dir_pseudo + '_' + atom_type + str(atom_num), base_path)
elif sim_prog == 'lammps':
    tde_run_path = os.path.relpath(lat_dir_pseudo + '_' + atom_type + str(atom_num) + '_lmp', base_path)
tde_run_outfile_path = os.path.relpath(os.path.join(tde_run_path, lat_dir_pseudo + '_' + atom_type + str(atom_num) + '_out.txt'), base_path)

# set atomic weights from POTCAR
mass_lines = subprocess.run(['grep', 'POMASS', os.path.join(inp_path, 'POTCAR')], capture_output = True, text = True)
mass_lines_list = mass_lines.stdout.split('\n ')
atomic_masses = []
for i in range(len(mass_lines_list)):
    atomic_masses.append(float(re.search(r'\d+.\d+', mass_lines_list[i]).group()))


# utility functions
def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    return vector / np.linalg.norm(vector)


def cart2sph(x, y, z):
    dxy = np.sqrt(x**2 + y**2)
    r = np.sqrt(dxy**2 + z**2)
    theta = np.arctan2(y, x)
    phi = np.arctan2(dxy, z)
    theta, phi = np.rad2deg([theta, phi])
    return r, theta % 360, phi


def sph2cart(theta, phi, r=0.5):
    theta, phi = np.deg2rad([theta, phi])
    z = r * np.cos(phi)
    rsinphi = r * np.sin(phi)
    x = rsinphi * np.cos(theta)
    y = rsinphi * np.sin(theta)
    return x, y, z


def vasp_inp_match_check(poscar_filepath, potcar_filepath):
    """
    Function to compare the order of elements in the POSCAR file to the order of
    elements in the POTCAR file and ensure the order matches.
    """
    # open POSCAR file and read lines into a list
    pos_f = open(poscar_filepath, 'r')
    pos_lines = pos_f.readlines()
    pos_f.close()
    
    # convert species names line from POSCAR file from Fortran to a list 
    pos_spec_line = ff.FortranRecordReader('(10A5)')  # arbitrarily choose max species number of 10
    pos_species = pos_lines[5]
    pos_spec_list = pos_spec_line.read(pos_species)
    for i in range(len(pos_spec_list)):
        pos_spec_list[i] = pos_spec_list[i].strip()
    pos_spec_list = list(filter(None, pos_spec_list)) # remove empty values from list
    
    # open POTCAR file and read lines into a list
    pot_f = open(potcar_filepath, 'r')
    pot_lines = pot_f.readlines()
    pot_f.close()
    
    # find POTCAR lines with species names
    pot_spec_lines = subprocess.run(['grep', 'TITEL', potcar_filepath], capture_output = True, text = True)
    pot_spec_lines_list = pot_spec_lines.stdout.split('\n ')
    pot_spec_list = []
    for i in range(len(pot_spec_lines_list)):
        pot_spec_list.append(re.split(r'[;,\s]\s*', pot_spec_lines_list[i])[4])
    
    if pos_spec_list == pot_spec_list:
        return True
    else:
        return False


# velocity vector calculation function
def lat_vel_calc(M, E, dir):
    """
    Function to calculate velocity vector of an atom in the lattice.
    Given an atomic mass (u), energy (eV), and crystallographic direction,
    returns the velocity vector for the lattice atom (Ang./fs).
    The direction and velocity are 1D numpy arrays.
    """
    M *= 931.49432*1e6    # 1/u * MeV/c^2 * eV/MeV
    
    v_mag = np.sqrt(np.divide(2*E, M))
    v_mag *= 299792458*1e10*1e-15    # m/s * Angstrom/m * s/fs
    
    v = np.zeros((v_mag.shape[0], dir.shape[0]))
    
    dir_mag = np.linalg.norm(dir)
    dir_unit_vec = dir/dir_mag
    
    for i in range(v_mag.shape[0]):
        v[i, :] = np.multiply(v_mag[i], dir_unit_vec)
    return v


# writing formatted velocity vector to POSCAR file
def vel_to_POSCAR(vel, atom_type, atom_no, filepath):
    """
    Function to write velocity vector of an atom in the lattice to the POSCAR file.
    Given a velocity vector (Ang./fs), atom type, and the number of the atom
    (from the POSCAR file). The velocity vector is a 1D numpy array, the atom
    type is a pymatgen Element, and the atom number is an integer.
    """
    # open POSCAR file and read lines into a list
    f = open(filepath, 'r')
    f_lines = f.readlines()
    f.close()
    
    # convert velocity vector into Fortran format
    vel_line = ff.FortranRecordWriter('(3E16.8)')
    vel_vector = vel_line.write(vel)+'\n'
    
    # convert species names line from POSCAR file from Fortran to a list 
    spec_line = ff.FortranRecordReader('(10A5)')  # arbitrarily choose max species number of 10
    species = f_lines[5]
    spec_list = spec_line.read(species)
    for i in range(len(spec_list)):
        spec_list[i] = spec_list[i].strip()
    spec_list = list(filter(None, spec_list)) # remove empty values from list
    
    # convert ions per species line from POSCAR file from Fortran to a list
    no_line = ff.FortranRecordReader('(10I6)')  # arbitrarily choose max species number of 10
    ion_nos = f_lines[6]
    no_list = no_line.read(ion_nos)
    no_list = list(filter(None, no_list)) # remove empty values from list
    total_ions = sum(no_list)
    
    # check if empty velocity lines exist, if not, add lines
    if len(f_lines) < 2*total_ions:
        if f_lines[-1].isspace() == False:
            f_lines += ['\n']
        f_lines += ['\n'] + [vel_line.write(np.array([0, 0, 0]))+'\n' for i in range(total_ions)]
    
    # replace velocity line for desired atom with converted velocity vector
    for i in range(len(spec_list)):           
        if atom_type.lower() in spec_list[i].lower():
            f_lines[8+total_ions+atom_no+sum(no_list[:i])] = vel_vector
    
    f = open(filepath, 'w+')
    f.writelines(f_lines)
    f.close()
    
    return


# read lattice directions and kinetic energies from text files
# finds lattice direction from pseudo passed from bash
lat_dir_list_file = os.path.relpath('latt_dirs_to_calc.csv', base_path)
with open(lat_dir_list_file) as ld_f:
    for line in ld_f:
        if line[:len(lat_dir_pseudo)] == lat_dir_pseudo:
            lat_dir_list = re.split(r'[;,\s]\s*', line.strip('\n'))  # [u v w] or (rho, phi, theta)
            break

kinE_list_file = os.path.relpath(os.path.join(tde_run_path, 'KE_calcs_list.txt'), base_path)
with open(kinE_list_file) as ke_f:
    for line in ke_f:
        pass
    kinE_line = line  # eV

kinE = np.array([float(kinE_line)])
print('KE:', kinE, 'eV')

tde_kinE_path = os.path.relpath(os.path.join(tde_run_path, kinE_line.strip('\n') + 'eV'), base_path)
pos_file = os.path.join(tde_kinE_path, 'POSCAR')
pot_file = os.path.join(tde_kinE_path, 'POTCAR')

# check if POTCAR and POSCAR are congruent
if vasp_inp_match_check(pos_file, pot_file) == False:
    raise ValueError

# open POSCAR file and read lines into a list
pos_f = open(pos_file, 'r')
pos_lines = pos_f.readlines()
pos_f.close()

# open POSCAR file and read lattice vector lines into a list
ai = [pos_lines[2][1:-1], pos_lines[3][1:-1], pos_lines[4][1:-1]]

# convert lattice vectors line from POSCAR file from Fortran to a list
lat_vec_line = ff.FortranRecordReader('(3F22.16)')
lat_vecs = np.array([lat_vec_line.read(ai[j]) for j in range(len(ai))])

# convert species names line from POSCAR file from Fortran to a list
pos_spec_line = ff.FortranRecordReader('(10A5)')
pos_species = pos_lines[5]
pos_spec_list = pos_spec_line.read(pos_species)
for i in range(len(pos_spec_list)):
    pos_spec_list[i] = pos_spec_list[i].strip()

if lat_dir_pseudo[-1] == 'L':
    lat_dir = np.array(lat_dir_list[5:], dtype=int)
    vel_dir = lat_dir@lat_vecs
    print('lattice direction:', str(lat_dir))
elif lat_dir_pseudo[-1] == 'S':
    lat_dir = np.array(lat_dir_list[5:], dtype=float)
    # rho can be any value since magnitude is scaled, choose 1 for simplicity
    # first angle (phi) needs to be the angle given in the heatmap - 60 deg since heatmap adds 60 deg
    cart_coords = np.array(sph2cart(float(lat_dir_list[5:][1]), float(lat_dir_list[5:][2]), r=1.))
    vel_dir = cart_coords
    print('spherical direction:', str(lat_dir))

# calculate velocity vector and write to POSCAR file
for i in range(len(pos_spec_list)):
    if atom_type.lower() == pos_spec_list[i].lower():
        vel_vecs = lat_vel_calc(atomic_masses[i], kinE, vel_dir)
vel_to_POSCAR(vel_vecs[0], atom_type, atom_num, pos_file)
print('velocity vectors:', vel_vecs)

out_file = open(tde_run_outfile_path, 'a')
out_file.write('direction:  '+str(lat_dir)+'\nvelocity vectors:  '+str(vel_vecs)+'\n')
out_file.close()



#!/usr/bin/env python3
"""Python module used to characterize defects resulting from TDE calculations."""
import os
import sys
import glob
import pprint
import fortranformat as ff

import math
import numpy as np
import pandas as pd

import warnings
warnings.filterwarnings('ignore', message='.*OVITO.*PyPI')

import ovito._extensions.pyscript

from ovito.io import *
from ovito.data import *
from ovito.modifiers import *
from ovito.pipeline import *


def find_contcars(base_path, pseudo='*'):
    """
    Function to create an array of post-CGM CONTCAR files for defect analysis.
    """
    contcar_files = np.array(glob.glob(os.path.join(base_path, pseudo, '*', 'cgm', 'CONTCAR'), recursive=True))
    return contcar_files


def find_dumpfiles(base_path, pseudo='*'):
    """
    Function to create an array of post-MD dump files for defect analysis.
    """
    dump_files = np.array(glob.glob(os.path.join(base_path, pseudo, '*', 'dump.final'), recursive=True))
    return dump_files


def pseudo_keys_from_file(pseudo_keyfile='pseudo_keys.csv'):
    """
    Function to convert a CSV file of pseudo keys to a dictionary.
    """
    pk_df = pd.read_csv(pseudo_keyfile, index_col=0)
    pks_dict = pk_df.to_dict('split')
    pks_dict = dict(zip(pks_dict['index'], pks_dict['data']))
    return pks_dict


def PBC_distance(x0, x1, dimensions):
    """
    Source: https://stackoverflow.com/a/11109336
    Function that takes 3 numpy arrays: an initial position array, a final position array,
    and the dimensions of the boundary. Returns an array for the distance, taking into
    account Periodic Boundary Conditions (PBC).
    """
    delta = np.abs(x0 - x1)
    delta = np.where(delta > 0.5 * dimensions, delta - dimensions, delta)
    return np.sqrt((delta ** 2).sum(axis=-1))


def cart2sph(x, y, z):
    dxy = np.sqrt(x**2 + y**2)
    r = np.sqrt(dxy**2 + z**2)
    theta = np.arctan2(y, x)
    phi = np.arctan2(dxy, z)
    theta, phi = np.rad2deg([theta, phi])
    return r, theta % 360, phi


def lat2sph(uvw, ai):
    """
    Function converting from lattice directions to spherical coordinates. Given two 2D
    arrays, one for lattice directions and one for lattice vectors, returns a 2D array
    of the spherical coordinates corresponding to the lattice directions.
    """
    xyz, rpt = np.zeros(uvw.shape), np.zeros(uvw.shape)
    
    for i in range(uvw.shape[0]):
        xyz[i, :] = uvw[i, :]@ai
        rpt[i, :] = np.around(cart2sph(xyz[i, 0], xyz[i, 1], xyz[i, 2]), decimals=2)
        rpt[i, 0] = 1.00
    
    return rpt


###########################################################################
######################## PYMATGEN DEFECTS ANALYSIS ########################
###########################################################################
def find_defects_multi_files(base_pos_path, cont_paths):
    """
    Given an initial POSCAR file and any number of post-AIMD, post-CGM CONTCAR files, returns the
    defect positions by comparing final atomic positions to initial atomic positions. File input
    is a string for the POSCAR filepath and an array of strings for the CONTCAR filepaths.
    """
    base_struct = Structure.from_file(base_pos_path)

    dsf = finder.DefectSiteFinder()
    defect_pos = np.zeros((cont_paths.shape[0], 3))
    
    for q in range(cont_paths.shape[0]):
        defect_struct = Structure.from_file(cont_paths[q])
        fpos = dsf.get_defect_fpos(defect_structure=defect_struct, base_structure=base_struct)
        fpos -= np.round(fpos)
        defect_pos[q, :] = fpos
    
    return defect_pos


def fp_sep_dist(base_pos_path, cont_paths, lattice_vecs, atom_type='ga', atom_no=34):
    """
    Given an initial POSCAR file and any number of post-AIMD, post-CGM CONTCAR files, returns the
    Frenkel pair separation distance after determining defect positions in the structure. File
    input is a string for the POSCAR filepath and an array of strings for the CONTCAR filepaths.
    """
    # this method assumes the vacancy is at the initial atomic position
    defect_pos = find_defects_multi_files(base_pos_path, cont_paths)@lattice_vecs
    v_pos = np.zeros((defect_pos.shape[0], 3))
    fp_sep = np.zeros((defect_pos.shape[0], 1))

    # open POSCAR file and read lines into a list
    pos_f = open(base_pos_path, 'r')
    pos_lines = pos_f.readlines()
    pos_f.close()

    # convert species names line from POSCAR file from Fortran to a list
    pos_spec_line = ff.FortranRecordReader('(2A5)')
    pos_species = pos_lines[5]
    pos_spec_list = pos_spec_line.read(pos_species)
    for i in range(len(pos_spec_list)):
        pos_spec_list[i] = pos_spec_list[i].strip()
    
    # convert ions per species line from POSCAR file from Fortran to a list
    pos_no_line = ff.FortranRecordReader('(2I6)')
    ion_nos = pos_lines[6]
    no_list = pos_no_line.read(ion_nos)
    total_ions = sum(no_list)
    
    # convert position line from POSCAR file from Fortran to a list
    pos_line = ff.FortranRecordReader('(3F20.16)')
    
    for i in range(len(pos_spec_list)):
        if atom_type.lower() in pos_spec_list[i].lower():
            atom_pos_line = 8+atom_no+sum(no_list[:i])
    
    for n in range(defect_pos.shape[0]):
        v_pos[n, :] = pos_line.read(pos_lines[atom_pos_line].strip('\n'))
        fp_sep[n] = math.dist(defect_pos[n, :], np.array([0, 0, 0]))
    
    return fp_sep

# print('######################## PYMATGEN DEFECTS ANALYSIS ########################')
# print(find_defects_multi_files(poscar_path, contcar_paths)@gan_lattice_vecs)
# print(fp_sep_dist(poscar_path, contcar_paths, gan_lattice_vecs, atom_type='ga', atom_no=34))

###########################################################################
############################ OVITO WS ANALYSIS ############################
###########################################################################
"""
how this needs to work
can probably just use a single function for everything, don't really have much repetition

WS analysis: output_displaced = False
- returns the vacant site location and the other atom closest to final interstitial
- will not know which is the vacant site yet
- can get the vacancy and interstitial count info here

WS analysis: output_displaced = True
- returns interstitial site location and other atom closest to final interstitial
- site that matches previous WS analysis can be thrown away, other site from first is v and other site from second is i
- need to choose a cutoff such that the additional site is either thrown away or considered as a dumbbell
- need for returning proper displacement vectors, so should be done second
- also need site type/index/identifier properties from this

Displacements modifier
- find displacement vectors, should be able to use the site index from second WS to determine final position of interstitial
- use this and site type/identifier to determine what type of interstitial and what position

final info of desire
v & i count (get directly from first WS)
v & i type: Ga/N (get directly from second WS?)
i site: t/o/db (disp mod and something else idk yet)
if v & i type are same (they should be, i think only case where they aren't is if there are multiple defects): separation distance (both WS and calculate distance formula)


main issues currently:    <---- this was for the previous version of this file, may not be relevant anymore
- problem regarding cutoff separating atoms inhabiting the lattice site closest to the interstitial being regarded as so vs. split dumbbell interstitial
- could be a problem with defining the array masks to delete rows (should they be the same? or could they be in different orders)
- this trickles down into the FP separation distance calculation
- need to use the displacement magnitude/vectors to determine whether interstitials are tetrahedral or octahedral
- also potential issues with runs like the [120] direction (17L?) where ovito recognizes 2 FPs (1 Ga, 1 N) as 1 FP
- need to bring in other info such as direction associated with pseudo (should be easy, use key_dict again)
"""
def defect_analysis(defect_file, ref_file, verbose=False):
    """
    Perform WS analyses
    """
    ws_comb_dict = {
        'sites': {},
        'atoms': {}
    }

    pipeline = import_file(defect_file)
    # displace atom config
    data = pipeline.compute()
    disp_type = data.particles['Particle Type'] 
    disp_pos = data.particles['Position'][...]

    # Perform Wigner-Seitz analysis (sites mode):
    ws = WignerSeitzAnalysisModifier(
        output_displaced = False,
        per_type_occupancies = True,
        affine_mapping = ReferenceConfigurationModifier.AffineMapping.ToReference,
        minimum_image_convention = True
        )
    
    # Load the reference config from a separate input file.
    ws.reference = FileSource() # Note: You may have to import FileSource from the ovito.pipeline module. 
    ws.reference.load(ref_file)
    pipeline.modifiers.append(ws)
    output_sites = pipeline.compute()


    def get_defects_from_WS(data):
        """
        Parse data gathered from Wigner-Seitz Analysis
        """
        ws_dict = {}

        occupancies = data.particles['Occupancy']
        site_pos = data.particles['Position'][...]

        # Get the site types as additional input:
        site_type = data.particles['Particle Type']
        # Calculate total occupancy of every site:
        total_occupancy = np.sum(occupancies, axis=1)

        ######## get vacancy information ########
        nAvac = np.sum((site_type == 1) & (total_occupancy == 0))
        nBvac = np.sum((site_type == 2) & (total_occupancy == 0))
        nvac = data.attributes['WignerSeitz.vacancy_count']
        print('\nIdenfity %d vacancies, including %d type1 and %d type2'%(nvac,nAvac,nBvac)) if verbose else None
        print('type1 vacancy site positions:') if verbose else None
        Avac_pos = site_pos[(site_type == 1) & (total_occupancy == 0)]
        print(Avac_pos) if verbose else None
        print('type2 vacancy site positions:') if verbose else None
        Bvac_pos = site_pos[(site_type == 2) & (total_occupancy == 0)]
        print(Bvac_pos) if verbose else None
        ws_dict['nAvac'], ws_dict['nBvac'], ws_dict['Avac_pos'], ws_dict['Bvac_pos'] = nAvac, nBvac, Avac_pos, Bvac_pos

        ######## get antisite information ########
        B2A = np.sum((site_type == 1) & (occupancies[:,1] == 1) & (total_occupancy == 1))
        A2B = np.sum((site_type == 2) & (occupancies[:,0] == 1) & (total_occupancy == 1))
        print('\nIdenfity %d antisites, including %d type1occupy2 and %d type2occupy1'%(B2A+A2B,A2B,B2A)) if verbose else None
        print('type2occupy1 site positions:') if verbose else None
        B2A_pos = site_pos[(site_type == 1) & (occupancies[:,1] == 1) & (total_occupancy == 1)]
        print(B2A_pos) if verbose else None
        print('type1occupy2 site positions:') if verbose else None
        A2B_pos = site_pos[(site_type == 2) & (occupancies[:,0] == 1) & (total_occupancy == 1)]
        print(A2B_pos) if verbose else None
        ws_dict['nB2A'], ws_dict['nA2B'], ws_dict['B2A_pos'], ws_dict['A2B_pos'] = B2A, A2B, B2A_pos, A2B_pos

        ######## get interstitial information ########
        int_sites = site_pos[total_occupancy > 1]
        int_site_types = site_type[total_occupancy > 1]
        finder = NearestNeighborFinder(np.max(total_occupancy), data)
        ws_dict['Aint_pos'], ws_dict['Bint_pos'] = [np.empty((0, 3), dtype=np.float64) for n in range(0, 2)]
        ws_dict['nAint'], ws_dict['nBint'] = 0, 0  # np.array([[]])
        for i,pos in enumerate(int_sites):
            print('\nIntersitial at site',pos,'with site type=',int_site_types[i]) if verbose else None
            if int_site_types[i] == 1 and pos not in ws_dict['Aint_pos']:
                ws_dict['Aint_pos'] = np.append(ws_dict['Aint_pos'], np.array([pos]), axis=0)
                ws_dict['nAint'] += 1
            elif int_site_types[i] == 2 and pos not in ws_dict['Bint_pos']:
                ws_dict['Bint_pos'] = np.append(ws_dict['Bint_pos'], np.array([pos]), axis=0)
                ws_dict['nBint'] += 1
            
            """ # determine nearest neighbors for each atom
            print('Atoms at this site')
            for neigh in finder.find_at(pos):
                ind = neigh.index
                print(disp_type[ind],disp_pos[ind])"""

        return ws_dict

    print('######## Wigner-Seitz Analysis (Sites) ########') if verbose else None
    ws_comb_dict['sites'] = get_defects_from_WS(output_sites)

    # Update Wigner-Seitz analysis (atoms mode):
    pipeline.modifiers[0].output_displaced = True
    output_atoms = pipeline.compute()

    print('######## Wigner-Seitz Analysis (Atoms) ########') if verbose else None
    ws_comb_dict['atoms'] = get_defects_from_WS(output_atoms)

    pprint.pprint(ws_comb_dict) if verbose else None

    # Compare both Wigner-Seitz analyses, return cross-analyzed data
    defect_dict = {
        'nAvac': int(ws_comb_dict['sites']['nAvac']),
        'nBvac': int(ws_comb_dict['sites']['nBvac']),
        'nA2B': int(ws_comb_dict['sites']['nA2B']),
        'nB2A': int(ws_comb_dict['sites']['nB2A']),
        'nAint': int(ws_comb_dict['atoms']['nAint']) - int(ws_comb_dict['sites']['nAint']),
        'nBint': int(ws_comb_dict['atoms']['nBint']) - int(ws_comb_dict['sites']['nBint']),
        'Avac_pos': ws_comb_dict['sites']['Avac_pos'],
        'Bvac_pos': ws_comb_dict['sites']['Bvac_pos'],
        'A2B_pos': ws_comb_dict['sites']['A2B_pos'],
        'B2A_pos': ws_comb_dict['sites']['B2A_pos'],
        'Aint_pos': ws_comb_dict['atoms']['Aint_pos'],  # np.empty((0, 3), dtype=np.float64),
        'Bint_pos': ws_comb_dict['atoms']['Bint_pos']  # np.empty((0, 3), dtype=np.float64)
    }

    # something is going wrong, here's some pseudocode
    """
    maybe set original lattice site either way and add to dict

    for each interstitial (use counts); need to
    1. determine distance from interstitial and displaced atom to original lattice site
        interstitial and displaced atom are both in defect_dict
        can both be in Aint_pos/Bint_pos, or have one in each
        original lattice site is in ws_comb_dict['sites']
        will be in same Aint_pos/Bint_pos based on location of displaced atom
    2a. if int&disp are in the same array, will need to determine which is interstitial based on which one is closer to original lattice site
    2b. if int&disp are in different arrays, maybe just treat the same?
    3. after determining int&disp, int will remain in interstitial list and disp will be categorized as...
        either nothing (and will be removed), if distance from site is < ~0.5-0.75 Ang
        or dumbbell config, if site is roughly shared, distance from site is ~ 0.75-1.75 Ang
    4. remove displaced atom either way (can therefore treat this the same, just adding dumbbell to dictionary either way)
        need to remove disp either way and add dumbbell to dict either way
        will just add a value if a dumbbell exists, will add empty if not
    5. use parsing function to add dumbbell into config if exists
    """

    r_i, r_ij = [], []
    og_A_latt_pos, disp_A_fin_pos, db_A_fin_pos = [np.empty((0, 3), dtype=np.float64) for n in range(0, 3)]
    og_B_latt_pos, disp_B_fin_pos, db_B_fin_pos = [np.empty((0, 3), dtype=np.float64) for n in range(0, 3)]
    for iA in range(defect_dict['nAint']):
        #og_A_latt_pos = og_A_latt_pos.append(ws_comb_dict['sites']['Aint_pos'][iA])
        pass
    for iB in range(defect_dict['nBint']):
        pass

    """r_i, r_ij = [], []
    disp_atom_fin_pos, db_atom_fin_pos = [], []
    if defect_dict['nAint'] < defect_dict['Aint_pos'].shape[0]:
        if defect_dict['nAint'] == 0:
            # Aint_pos contains only displaced atom, Bint_pos contains interstitial
            for i in range(defect_dict['Aint_pos'].shape[0]):
                r_i.append(defect_dict['Aint_pos'][i])
                for j in range(ws_comb_dict['sites']['Aint_pos'].shape[0]):
                    r_ij.append(PBC_distance(defect_dict['Aint_pos'][i], ws_comb_dict['sites']['Aint_pos'][j], gan_params))
            # need to find index of minimum r_ij value corresponding to first set of r_i values
            # aka need to slice r_ij list based on [(i*ws_comb_dict['sites']['Aint_pos'].shape[0]):(i+1)*ws_comb_dict['sites']['Aint_pos'].shape[0])]
            for i in range(len(r_i)):
                j_start = i*ws_comb_dict['sites']['Aint_pos'].shape[0]
                j_end = (i + 1)*ws_comb_dict['sites']['Aint_pos'].shape[0]
                j_min = np.argmin(r_ij[j_start:j_end]) + j_start
                if r_ij[j_min] <= 0.75:  # small displacement
                    disp_atom_fin_pos.append(r_i[i])
                elif r_ij[j_min] > 0.75 and r_ij[j_min] <= 1.75:  # dumbbell config
                    db_atom_fin_pos.append(r_i[i])
                else:
                    pass
                defect_dict['Aint_pos'] = np.delete(defect_dict['Aint_pos'], r_i[i])
        else:
            # Aint_pos contains both displaced atom and interstitial
            # need to break in case of interstitial
            for i in range(defect_dict['Aint_pos'].shape[0]):
                r_i.append(defect_dict['Aint_pos'][i])
                for j in range(ws_comb_dict['sites']['Aint_pos'].shape[0]):
                    r_ij.append(PBC_distance(defect_dict['Aint_pos'][i], ws_comb_dict['sites']['Aint_pos'][j], gan_params))
            for i in range(len(r_i)):
                j_start = i*ws_comb_dict['sites']['Aint_pos'].shape[0]
                j_end = (i + 1)*ws_comb_dict['sites']['Aint_pos'].shape[0]
                j_min = np.argmin(r_ij[j_start:j_end]) + j_start
                if r_ij[j_min] <= 0.75:  # small displacement
                    disp_atom_fin_pos.append(r_i[i])
                elif r_ij[j_min] > 0.75 and r_ij[j_min] <= 1.75:  # dumbbell config
                    db_atom_fin_pos.append(r_i[i])
                else:
                    pass
                defect_dict['Aint_pos'] = np.delete(defect_dict['Aint_pos'], r_i[i])
    elif defect_dict['nBint'] < defect_dict['Bint_pos'].shape[0]:
        if defect_dict['nBint'] == 0:
            # Bint_pos contains only displaced atom, Aint_pos contains interstitial
            for i in range(defect_dict['Bint_pos'].shape[0]):
                r_i.append(defect_dict['Bint_pos'][i])
                for j in range(ws_comb_dict['sites']['Bint_pos'].shape[0]):
                    r_ij.append(PBC_distance(defect_dict['Bint_pos'][i], ws_comb_dict['sites']['Bint_pos'][j], gan_params))
            for i in range(len(r_i)):
                j_start = i*ws_comb_dict['sites']['Bint_pos'].shape[0]
                j_end = (i + 1)*ws_comb_dict['sites']['Bint_pos'].shape[0]
                j_min = np.argmin(r_ij[j_start:j_end]) + j_start
                if r_ij[j_min] <= 0.75:  # small displacement
                    disp_atom_fin_pos.append(r_i[i])
                elif r_ij[j_min] > 0.75 and r_ij[j_min] <= 1.75:  # dumbbell config
                    db_atom_fin_pos.append(r_i[i])
                else:
                    pass
                defect_dict['Bint_pos'] = np.delete(defect_dict['Bint_pos'], r_i[i])
        else:
            # Bint_pos contains both displaced atom and interstitial
            for i in range(defect_dict['Bint_pos'].shape[0]):
                r_i.append(defect_dict['Bint_pos'][i])
                for j in range(ws_comb_dict['sites']['Bint_pos'].shape[0]):
                    r_ij.append(PBC_distance(defect_dict['Bint_pos'][i], ws_comb_dict['sites']['Bint_pos'][j], gan_params))
            for i in range(len(r_i)):
                j_start = i*ws_comb_dict['sites']['Bint_pos'].shape[0]
                j_end = (i + 1)*ws_comb_dict['sites']['Bint_pos'].shape[0]
                j_min = np.argmin(r_ij[j_start:j_end]) + j_start
                if r_ij[j_min] <= 0.75:  # small displacement
                    disp_atom_fin_pos.append(r_i[i])
                elif r_ij[j_min] > 0.75 and r_ij[j_min] <= 1.75:  # dumbbell config
                    db_atom_fin_pos.append(r_i[i])
                else:
                    pass
                defect_dict['Bint_pos'] = np.delete(defect_dict['Bint_pos'], r_i[i])"""

    # can use now accurate Aint & Bint values
    # if either Aint or Bint array is longer than the nAint or nBint, from array that is longer, determine which value is closest to original site position
    # set this closer position to a different variable, remove from int_pos array
    # if distance is > 0.5ish, add dumbbell position to dictionary
    """lattice_site_org_pos, disp_atom_fin_pos, db_atom_fin_pos = [], [], []
    for i in range(ws_comb_dict['atoms']['Aint_pos'].shape[0]):
        for j in range(ws_comb_dict['sites']['Aint_pos'].shape[0]):
            print(ws_comb_dict['atoms']['Aint_pos'][i], ws_comb_dict['sites']['Aint_pos'][j])
            print(PBC_distance(ws_comb_dict['atoms']['Aint_pos'][i], ws_comb_dict['sites']['Aint_pos'][j], gan_params))

            if (ws_comb_dict['atoms']['Aint_pos'][i] == ws_comb_dict['sites']['Aint_pos'][j]).all():
                lattice_site_org_pos.append(ws_comb_dict['atoms']['Aint_pos'][i])
            elif PBC_distance(ws_comb_dict['atoms']['Aint_pos'][i], ws_comb_dict['sites']['Aint_pos'][j], gan_params) <= 0.75 and PBC_distance(ws_comb_dict['atoms']['Aint_pos'][i], ws_comb_dict['sites']['Aint_pos'][j], gan_params) > 0.:
                disp_atom_fin_pos.append(ws_comb_dict['sites']['Aint_pos'][i])
            elif PBC_distance(ws_comb_dict['atoms']['Aint_pos'][i], ws_comb_dict['sites']['Aint_pos'][j], gan_params) <= 1.0 and PBC_distance(ws_comb_dict['atoms']['Aint_pos'][i], ws_comb_dict['sites']['Aint_pos'][j], gan_params) > 0.75:
                db_atom_fin_pos.append(ws_comb_dict['sites']['Aint_pos'][i])
            else:
                defect_dict['Aint_pos'] = ws_comb_dict['atoms']['Aint_pos']
    
    for i in range(ws_comb_dict['atoms']['Bint_pos'].shape[0]):
        for j in range(ws_comb_dict['sites']['Bint_pos'].shape[0]):
            print(ws_comb_dict['atoms']['Bint_pos'][i], ws_comb_dict['sites']['Bint_pos'][j])
            print(PBC_distance(ws_comb_dict['atoms']['Bint_pos'][i], ws_comb_dict['sites']['Bint_pos'][j], gan_params))

            if (ws_comb_dict['atoms']['Bint_pos'][i] == ws_comb_dict['sites']['Bint_pos'][j]).all():
                lattice_site_org_pos.append(ws_comb_dict['atoms']['Bint_pos'][i])
            elif PBC_distance(ws_comb_dict['atoms']['Bint_pos'][i], ws_comb_dict['sites']['Bint_pos'][j], gan_params) <= 0.75 and PBC_distance(ws_comb_dict['atoms']['Bint_pos'][i], ws_comb_dict['sites']['Bint_pos'][j], gan_params) > 0.:
                disp_atom_fin_pos.append(ws_comb_dict['sites']['Bint_pos'][i])
            elif PBC_distance(ws_comb_dict['atoms']['Bint_pos'][i], ws_comb_dict['sites']['Bint_pos'][j], gan_params) <= 1.0 and PBC_distance(ws_comb_dict['atoms']['Bint_pos'][i], ws_comb_dict['sites']['Bint_pos'][j], gan_params) > 0.75:
                db_atom_fin_pos.append(ws_comb_dict['sites']['Bint_pos'][i])
            else:
                defect_dict['Bint_pos'] = ws_comb_dict['atoms']['Bint_pos']
    
    defect_dict['nAint'], defect_dict['nBint'] = defect_dict['Aint_pos'].shape[0], defect_dict['Bint_pos'].shape[0]"""

    #print(og_atom_latt_pos, disp_atom_fin_pos, db_atom_fin_pos)

    return defect_dict


######################################
# function to determine TDE from a single datafile
######################################
def get_tde_from_datafile(data_filepath='*_data.csv', e_tol=1.0, ke_cut=45):
    """
    Scans a findTDE *_data.csv file and returns the TDE value for the specified knockout type.
    """
    # read in tde data into structured array
    tde_data = np.genfromtxt(data_filepath, delimiter=', ', skip_header=1,
                             dtype=[('pseudo', '<U4'), ('atom_type', '<U4'), ('atom_num', '<i4'),
                                ('ke', '<i4'), ('e_fin', '<f8'), ('dE', '<f8')]
                            )
    
    # sort array from lowest KE to highest KE
    tde_data = np.sort(tde_data, order='ke')
    
    # find TDE value from energy difference
    # first KE corresponding to a dE > e_tol is the TDE in the sorted array
    tde_value = 0
    for i in range(tde_data.shape[0]):
        if np.isnan(tde_data[i]['dE']) == True:
            continue
        elif tde_data[i]['dE'] > e_tol:
            if tde_data[i]['ke'] <= ke_cut:
                tde_value = tde_data[i]['ke']
            elif tde_data[i]['ke'] > ke_cut:
                tde_value = ke_cut
            break
        else:
            tde_value = ke_cut
    
    return tde_value


######################################
# function to average TDE values for each atom type for a given direction range
# function currently assumes rad-induced defect config is Frenkel pair
######################################
def average_tde_vals(defects_dict, pseudo_keys, lattice_vecs, dir_range=((0, 360), (0, 180))):
    """
    Function to average the TDE values for each atom type over the given range of directions.
    """
    tde_averages = {
        'avg': {
            'E_d_avg_A': [],
            'E_d_avg_B': []
        },
        'std': {
            'E_d_std_A': [],
            'E_d_std_B': []
        },
        'weights': {
            'E_d_wi_A': [],
            'E_d_wi_B': []
        }
    }
    
    cnt_A_FP, cnt_B_FP, cnt_complex, cnt_oor, cnt_cutoff = 0, 0, 0, 0, 0
    # need variable dtheta, dphi
    dtheta, dphi = np.deg2rad(7.5), np.deg2rad(7.5)

    for key in defects_dict.keys():
        # check the direction and pass if outside of range
        if key[-1] == 'S':
            phi, theta = pseudo_keys[key][0], pseudo_keys[key][1]
        elif key[-1] == 'L':
            phi, theta = lat2sph(np.array([pseudo_keys[key]]), lattice_vecs)[0][1], lat2sph(np.array([pseudo_keys[key]]), lattice_vecs)[0][2]
        else:
            raise ValueError('Incorrect direction pseudo.')
        
        if (phi < dir_range[0][0]) or (phi > dir_range[0][1]) or (theta < dir_range[1][0]) or (theta > dir_range[1][1]):
            cnt_oor += 1
            continue

        # check the resultant defect type
        if (defects_dict[key]['nAint'] > 0) and (defects_dict[key]['nAvac'] > 0):
            defect_type = 'A'
            cnt_A_FP += 1
        elif (defects_dict[key]['nBint'] > 0) and (defects_dict[key]['nBvac'] > 0):
            defect_type = 'B'
            cnt_B_FP += 1
        else:
            #print('{} --> Other type: nAint = {}, nAvac = {}, nBint = {}, nBvac = {}'.format(key, defects_dict[key]['nAint'], defects_dict[key]['nAvac'], defects_dict[key]['nBint'], defects_dict[key]['nBvac']))
            #print(defects_dict[key])
            cnt_complex += 1
            continue
        
        # categorize TDE values by the defect type
        tde_averages['avg'][f'E_d_avg_{defect_type}'].append(defects_dict[key]['E_d'])
        tde_averages['weights'][f'E_d_wi_{defect_type}'].append(np.sin(np.deg2rad(theta))*dtheta*dphi)
    
    # find standard deviation of the TDE values for each defect type
    for key in tde_averages['std'].keys():
        tde_averages['std'][key] = np.std(tde_averages['avg'][key.replace('std', 'avg')], dtype=np.float64)
    
    # average the TDE values for each defect type
    for key in tde_averages['avg'].keys():
        # standard average
        # tde_averages['avg'][key] = np.mean(tde_averages['avg'][key], dtype=np.float64)
        
        # weighted average
        tde_averages['avg'][key] = np.average(tde_averages['avg'][key], weights=tde_averages['weights'][key.replace('avg', 'wi')])
    del tde_averages['weights']
    
    print('A FPs:', cnt_A_FP, '\tB FPs:', cnt_B_FP, '\tOther:', cnt_complex, '\tOut of range:', cnt_oor)

    return tde_averages


######################################
# can then take the code i started to make previously to analyze this info
# maybe need to change how it is stored first in the original function
######################################
def parse_defect_properties(defects_dict, atom_types=['Ga', 'N']):
    """
    Function to analyze the defect properties for each calculation contained
    in a dictionary.
    """

    # create initial DataFrame to emulate table
    defects_df = pd.DataFrame(data=defects_dict)
    
    # add empty rows to Dataframe for additional data
    defects_df.loc['Defect Type'] = pd.Series([None for n in range(defects_df.columns.shape[0])])
    defects_df.loc['d_sep'] = pd.Series([None for n in range(defects_df.columns.shape[0])])

    # iterate through columns (calculation directions)
    for i in range(defects_df.columns.shape[0]):
        vacancies_list = ['V_{}'.format(atom_types[0]) for v in range(defects_df[defects_df.columns[i]].loc['nAvac'])] + \
            ['V_{}'.format(atom_types[1]) for v in range(defects_df[defects_df.columns[i]].loc['nBvac'])]
        antisites_list = ['{}_{}'.format(atom_types) for a in range(defects_df[defects_df.columns[i]].loc['nA2B'])] + \
            ['{}_{}'.format(atom_types[::-1]) for a in range(defects_df[defects_df.columns[i]].loc['nB2A'])]
        interstitials_list = ['{}_i'.format(atom_types[0]) for t in range(defects_df[defects_df.columns[i]].loc['nAint'])] + \
            ['{}_i'.format(atom_types[1]) for t in range(defects_df[defects_df.columns[i]].loc['nBint'])]
        defects_df[defects_df.columns[i]].loc['Defect Type'] = ' + '.join(vacancies_list+antisites_list+interstitials_list)
        # could determine O or T by having a list of possible O and T sites, determine distance from each via displacement vector, choose lowest distance

        if defects_df[defects_df.columns[i]].loc['nAvac'] == 1 and defects_df[defects_df.columns[i]].loc['nAint'] == 1:
            defects_df[defects_df.columns[i]].loc['d_sep'] = PBC_distance(defects_df[defects_df.columns[i]].loc['Aint_pos'], defects_df[defects_df.columns[i]].loc['Avac_pos'], gan_params)
        elif defects_df[defects_df.columns[i]].loc['nBvac'] == 1 and defects_df[defects_df.columns[i]].loc['nBint'] == 1:
            defects_df[defects_df.columns[i]].loc['d_sep'] = PBC_distance(defects_df[defects_df.columns[i]].loc['Bint_pos'], defects_df[defects_df.columns[i]].loc['Bvac_pos'], gan_params)

    # remove unnecessary info used in analysis for the purposes of displaying
    defects_df = defects_df.drop(['nAvac', 'nBvac', 'nA2B', 'nB2A', 'nAint', 'nBint'])
    
    # sort DataFrame
    defects_df.sort_index(inplace=True)

    # would be good to include information such as the knockout atom and the Ed value

    return defects_df


#####command: ovitos xx.py
if __name__ == "__main__":
    FILES_LOC = os.getcwd()    # os.path.dirname(__file__)
    
    run_type, knockout_atom = 'vasp', 'ga34'

    if run_type == 'vasp':
        defect_files = find_contcars(FILES_LOC, pseudo=f'*{knockout_atom}*')
    elif run_type == 'lammps':
        defect_files = find_dumpfiles(FILES_LOC, pseudo=f'*{knockout_atom}*')    # test with 10*S_ga34*
    # pprint.pprint(defect_files)
    
    # read in VASP and LAMMPS perfect structure information
    if run_type == 'vasp':
        vasp_ref_file = os.path.join(FILES_LOC, 'inp', 'POSCAR')
    elif run_type == 'lammps':
        vasp_ref_file, lmp_ref_file = os.path.join(FILES_LOC, 'inp', 'POSCAR'), os.path.join(FILES_LOC, 'inp', 'read_data_perfect.lmp')
    pos_f = open(vasp_ref_file, 'r')
    pos_lines = pos_f.readlines()
    pos_f.close()

    # open POSCAR file and read lattice vector lines into a list
    ai = [pos_lines[2][1:-1], pos_lines[3][1:-1], pos_lines[4][1:-1]]

    # convert lattice vectors line from POSCAR file from Fortran to a list
    lat_vec_line = ff.FortranRecordReader('(3F22.16)')
    vasp_lattice_vecs = np.array([lat_vec_line.read(ai[j]) for j in range(len(ai))])
    vasp_params = np.array([np.linalg.norm(vasp_lattice_vecs[0, :]), np.linalg.norm(vasp_lattice_vecs[1, :]), np.linalg.norm(vasp_lattice_vecs[2, :])])

    # add analysis information for each direction to a dict
    defect_configs_dict = {}

    for i in range(defect_files.shape[0]):
        if run_type == 'vasp':
            calc_desc, calc_tde = defect_files[i].split('/')[-4], int(defect_files[i].split('/')[-3][:-2])
            calc_pseudo, calc_atom_info = calc_desc.split('_')[0], calc_desc.split('_')[1]
            
            # get TDE value for direction from data file
            pseudo_datafile = os.path.abspath(os.path.join(defect_files[i], os.pardir, os.pardir, os.pardir, calc_pseudo+'_'+calc_atom_info+'_data.csv'))
            pseudo_tde_value = get_tde_from_datafile(pseudo_datafile, e_tol=1.0, ke_cut=45)
        elif run_type == 'lammps':
            calc_desc, calc_tde = defect_files[i].split('/')[-3], int(defect_files[i].split('/')[-2][:-2])
            calc_pseudo, calc_atom_info = calc_desc.split('_')[0], calc_desc.split('_')[1]
            
            # get TDE value for direction from data file
            pseudo_datafile = os.path.abspath(os.path.join(defect_files[i], os.pardir, os.pardir, calc_pseudo+'_'+calc_atom_info+'_data.csv'))
            pseudo_tde_value = get_tde_from_datafile(pseudo_datafile, e_tol=1.0, ke_cut=100)
        
        # print(calc_desc, ': ', calc_tde, ' eV\t (TDE: ', pseudo_tde_value, ' eV)' sep='')
        
        # perform defect analysis
        if calc_tde == pseudo_tde_value:
            defect_configs_dict[calc_pseudo] = defect_analysis(defect_files[i], vasp_ref_file)
            defect_configs_dict[calc_pseudo]['E_d'] = pseudo_tde_value
        elif calc_tde != pseudo_tde_value:
            # potentially need to add cutoff handling here
            if pseudo_tde_value >= 45:
                print('cutoff needed')
            continue
        else:
            continue

    print('############################ OVITO WS ANALYSIS SUMMARY ############################')
    # pprint.pprint(defect_configs_dict)
    if run_type == 'vasp':
        vasp_pseudo_keys = pseudo_keys_from_file(pseudo_keyfile=os.path.join(FILES_LOC, 'gan_vasp_pseudo_keys.csv'))
    elif run_type == 'lammps':
        lmp_pseudo_keys = pseudo_keys_from_file(pseudo_keyfile=os.path.join(FILES_LOC, 'gan_lmp_pseudo_keys.csv'))
    pprint.pprint(average_tde_vals(defect_configs_dict, vasp_pseudo_keys, vasp_lattice_vecs, dir_range=((30, 150), (0, 180))))

    # defect_configs_df = parse_defect_properties(defect_configs_dict, atom_types=['Ga', 'N']).T
    # pprint.pprint(defect_configs_df)



#!/usr/bin/env python3
"""Python module to convert VASP POSCAR to LAMMPS read_data file"""
import numpy as np
import os
import sys
import itertools


def pbc_distance(r1,r2,dimensions):
    # periodic distance given two points; dimensions is a numpy array, 
    delta = np.abs(r1 - r2)
    delta = np.where(delta > 0.5 * dimensions, delta - dimensions, delta)
    return np.sqrt((delta ** 2).sum(axis=-1))


#https://docs.lammps.org/Howto_triclinic.html
def convert_basis(A,B,C):
    ax = np.linalg.norm(A)
    bx = np.dot(B,A/ax)
    by = np.linalg.norm(np.cross(A/ax,B))
    cx = np.dot(C,A/ax)
    cy = (np.dot(B,C)-bx*cx)/by
    cz = np.sqrt(np.dot(C,C)-cx**2-cy**2)
    return np.array([[ax,bx,cx],[0,by,cy],[0,0,cz]])


class poscar2lmp:
    def __init__(self,posfile,outfile):
        self.read_pos(posfile)
        self.write_data(outfile)

    def read_pos(self,posfile):
        f = open(posfile,'r')
        headerlines = f.readlines()

        elemsline = 5
        dimlines = [2,3,4]
        atomsline = elemsline + 1
        headerline = 8
        if all(map(str.isdigit,headerlines[elemsline].strip().split()[0])): # if no element names exist
            atomsline = elemsline 
            headerline = 7

        scaling = float(headerlines[1])
        self.dims = []
        self.los = np.zeros(3)
        self.his = np.zeros(3)
        for i,item in enumerate(dimlines):
            lin = headerlines[item].strip().split()
            self.dims.append([float(k) for k in lin])
        self.dims = np.array(self.dims)
        atom_counts = [int(i) for i in headerlines[atomsline].strip().split()]
        self.natom = np.sum(atom_counts)
        f.close()

        data = [] #position data 
        k = headerline
        for line in headerlines[headerline:]:
             lin = line.strip().split()
             k += 1
             if len(lin) == 0:
                 continue
             data.append([float(i) for i in lin[:3]])
             if len(data) == self.natom:
                 break
        data = np.array(data)
        if headerlines[headerline-1].strip().split()[0] in ['direct','Direct','D']:
            data = np.dot(data,self.dims)*scaling
        data_v = [] #velocity data
        for line in headerlines[k:]:
             lin = line.strip().split()
             if len(lin) == 0:
                 continue
             data_v.append([float(i) for i in lin[:3]])
             if len(data_v) == self.natom:
                 break
        data_v = np.array(data_v)*1000  # scale to Ang/ps
        
        A,B,C = self.dims 
        Vol = np.linalg.norm(np.dot(np.cross(A,B),C))
        abc = convert_basis(A,B,C)
        factor = 1/Vol*np.dot(abc,np.array([np.cross(B,C),np.cross(C,A),np.cross(A,B)]))
        pos = np.dot(data,factor.transpose())
        vel = np.dot(data_v,factor.transpose())
        self.his[0] = abc[0,0] 
        self.his[1] = abc[1,1]
        self.his[2] = abc[2,2]
        if abs(abc[0,1]) <= abs((self.his[0] - self.los[0])/2.):
            self.xy_xz_yz = [abc[0,1],abc[0,2],abc[1,2]]
        elif abs(abc[0,1]) > abs((self.his[0] - self.los[0])/2.) and abc[0,1] < 0.:
            abc[0,1] += abs(self.his[0] - self.los[0])
            self.xy_xz_yz = [abc[0,1],abc[0,2],abc[1,2]]
        elif abs(abc[0,1]) > abs((self.his[0] - self.los[0])/2.) and abc[0,1] > 0.:
            abc[0,1] -= abs(self.his[0] - self.los[0])
            self.xy_xz_yz = [abc[0,1],abc[0,2],abc[1,2]]


        atomids = np.arange(1,self.natom+1)
        atomtypes = np.concatenate([np.ones(item)*(i+1) for i,item in enumerate(atom_counts)])
        self.atom_types = len(atom_counts)
        self.data = np.hstack((atomids.reshape(-1,1),atomtypes.reshape(-1,1),pos))
        self.data_v = np.hstack((atomids.reshape(-1,1),vel))


    def write_data(self,outfile):
        # write to readdata for lammps
            f = open(outfile,'w')
            f.write('Generate lammps data file from poscar\n\n')
            f.write('%d  atoms\n' %(self.natom))
            f.write('%d  %s\n\n' %(self.atom_types,'atom types'))
            f.write('%.8e  %.8e  %s\n' %(self.los[0],self.his[0],'xlo xhi'))
            f.write('%.8e  %.8e  %s\n' %(self.los[1],self.his[1],'ylo yhi'))
            f.write('%.8e  %.8e  %s\n' %(self.los[2],self.his[2],'zlo zhi'))
            f.write('%.8e  %.8e  %.8e xy xz yz\n\n' %(tuple(self.xy_xz_yz)))
            f.write('%s\n\n' %('Atoms'))
            for d in self.data:
                f.write('%d %d %.8e %.8e %.8e\n' %(tuple(d)))

            f.write('\nVelocities\n\n')
            for d in self.data_v:
                f.write('%d %.8e %.8e %.8e\n' %(tuple(d)))
            f.close()


###### IF XY IS TOO LARGE, ADD XHI TO IT. SIMILIAR TREATMENT FOR XZYZ I THINK
if __name__ == '__main__':
    posfile = os.path.join(os.getcwd(), 'POSCAR')
    outfile = os.path.join(os.getcwd(), 'read_data.lmp')
    poscar2lmp(posfile,outfile)
    
